"""Read Data Set."""
import pickle
import click
# import json
import pandas as pd
# import numpy as np
# from sklearn.linear_model import LinearRegression


@click.command()
@click.argument('y3', type=str)  # y3 is the input argument for path to dataset
def reader(y3):
    """Reader."""
    dataset = pd.read_csv(y3)

    print("this dataset", dataset)
    dataset['experience'].fillna(0, inplace=True)
    dataset['test_score'].fillna(dataset['test_score'].mean(), inplace=True)
    # Extracting rows (X-> independent variables and Y-> dependent/target variable)

    X = dataset.iloc[:, :3]   # Extracting first three columns from the dataset
    print('My X', X)

    Y = dataset.iloc[:, -1]  # Extracting last column from the dataset for target variable

    # Exporting X and Y as pickle files on to the disk
    pickle.dump(X, open("MyX.p", "wb"))
    pickle.dump(Y, open("MyY.p", "wb"))
    df1 = X.to_json()
    df2 = Y.to_json()


if __name__ == '__main__':
    reader(y3="")
# Ignores preserving code for now
# pylama:ignore=C0103,R1732,W0612



"""Workflow endpoints for getting metadata."""

from flask import make_response, jsonify
from flask_restful import Resource
from beeflow.wf_manager.common import wf_db


class WFMetadata(Resource):
    """Class for getting metadata."""

    def get(self, wf_id):
        """Get and return metadata."""
        # For now, just get the bolt port. Later we may want to add more information here.
        bolt_port = wf_db.get_bolt_port(wf_id)
        metadata = {
            'bolt_port': bolt_port,
        }
        return make_response(jsonify(metadata), 200)



"""Manage actions for tasks coming from the WFM."""
import traceback
from flask import jsonify, make_response
from flask_restful import Resource
from beeflow.common import log as bee_logging
from beeflow.task_manager import utils

log = bee_logging.setup(__name__)


class TaskActions(Resource):
    """Actions to take for tasks."""

    @staticmethod
    def delete():
        """Cancel received from WFM to cancel job, update queue to monitor state."""
        db = utils.connect_db()
        worker = utils.worker_interface()
        cancel_msg = ""
        for job in db.job_queue:
            task_id = job.task.id
            job_id = job.job_id
            name = job.task.name
            log.info(f"Cancelling {name} with job_id: {job_id}")
            try:
                job_state = worker.cancel_task(job_id)
            except Exception as err: # noqa (we have to catch everything here)
                log.error(err)
                log.error(traceback.format_exc())
                job_state = 'ZOMBIE'
            cancel_msg += f"{name} {task_id} {job_id} {job_state}"
        db.job_queue.clear()
        db.submit_queue.clear()
        resp = make_response(jsonify(msg=cancel_msg, status='ok'), 200)
        return resp



"""Container build utility code."""
import jsonpickle


def arg2task(task_arg):
    """Convert JSON encoded task to Task object.

    The build driver will expect a Task object, and the build
    interface starts with a JSON representation of the Task object.
    """
    return jsonpickle.decode(task_arg)


def task2arg(task):
    """Convert Task object to JSON encoded string.

    The build interface needs to pass Task data on the command line,
    because each compute node needs to understand the Task description.
    JSON format is a convenient way to describe the Task object at the
    command line.
    """
    return jsonpickle.encode(task)


class ContainerBuildError(Exception):
    """Cotnainer build error class."""



#!/usr/bin/env python3

"""beeflow.

This script manages the startup of the BEE daemons and supporting services.
If no arguments are given this script will start the BEEWorkflowManager,
BEETaskManager, and all required supporting services. If any combination of
services is specified using the appropriate flag(s) then ONLY those services
will be started.
"""
import os
import signal
import subprocess
import socket
import sys
import shutil
import datetime
import time

import daemon
import typer

from beeflow.client import bee_client
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common import cli_connection
from beeflow.common import paths
from beeflow.wf_manager.resources import wf_utils

from beeflow.common.db import wfm_db
from beeflow.common.db.bdb import connect_db
from beeflow.wf_manager.common import dep_manager


class ComponentManager:
    """Component manager class."""

    def __init__(self):
        """Construct the component manager."""
        self.components = {}
        self.procs = {}

    def component(self, name, deps=None):
        """Return a decorator function to be called."""

        def wrap(fn):
            """Add the component to the list."""
            self.components[name] = {
                'fn': fn,
                'deps': deps,
                'restart_count': 0,
                'failed': False,
            }

        return wrap

    def _validate(self, base_components):
        """Make sure that the components all exist and have valid deps."""
        missing = [name for name in base_components if name not in self.components]
        for component in self.components.values():
            if component['deps'] is None:
                continue
            for dep in component['deps']:
                if dep not in self.components:
                    missing.append(dep)
        if missing:
            raise RuntimeError(f'Missing/unknown component(s): {",".join(missing)}')

    def _find_order(self, base_components):
        """Find the order of the dependencies to launch."""
        s = base_components[:]
        levels = {name: 0 for name in self.components}
        while len(s) > 0:
            name = s.pop()
            if self.components[name]['deps'] is None:
                continue
            for dep in self.components[name]['deps']:
                levels[dep] = max(levels[name] + 1, levels[dep])
                # Detect a possible cycle
                if levels[dep] > len(self.components):
                    raise RuntimeError(f'There may be a cycle for the "{dep}" component')
                s.append(dep)
        levels = list(levels.items())
        levels.sort(key=lambda t: t[1], reverse=True)
        return [name for name, level in levels]

    def run(self, base_components):
        """Start and run everything."""
        # Determine if there are any missing components listed
        self._validate(base_components)
        # Determine the order to launch components in (note: this should just ignore cycles)
        order = self._find_order(base_components)
        print(f'Launching components in order: {order}')
        # Now launch the components
        for name in order:
            component = self.components[name]
            self.procs[name] = component['fn']()

    def poll(self):
        """Poll each process to check for errors, restart failed processes."""
        # Max number of times a component can be restarted
        max_restarts = bc.get('DEFAULT', 'max_restarts')
        for name in self.procs:  # noqa no need to iterate with items() since self.procs may be set
            component = self.components[name]
            if component['failed']:
                continue
            returncode = self.procs[name].poll()
            if returncode is not None:
                log = paths.log_fname(name)
                print(f'Component "{name}" failed, check log "{log}"')
                if component['restart_count'] >= max_restarts:
                    print(f'Component "{name}" has been restarted {max_restarts} '
                          'times, not restarting again')
                    component['failed'] = True
                else:
                    restart_count = component['restart_count']
                    print(f'Attempting restart {restart_count} of "{name}"...')
                    self.procs[name] = component['fn']()
                    component['restart_count'] += 1

    def status(self):
        """Return the statuses for each process in a dict."""
        return {
            name: 'RUNNING' if proc.poll() is None else 'FAILED'
            for name, proc in self.procs.items()
        }

    def kill(self):
        """Kill all components."""
        for name, proc in self.procs.items():
            print(f'Killing {name}')
            proc.terminate()


def warn(*pargs):
    """Print a red warning message."""
    typer.secho(' '.join(pargs), fg=typer.colors.RED, file=sys.stderr)


def launch_with_gunicorn(module, sock_path, *args, **kwargs):
    """Launch a component with Gunicorn."""
    # Setting the timeout to infinite, since sometimes the gdb can take too long
    return subprocess.Popen(['gunicorn', module, '--timeout', '0', '-b', f'unix:{sock_path}'],
                            *args, **kwargs)


def open_log(component):
    """Determine the log for the component, open and return it."""
    log = paths.log_fname(component)
    return open(log, 'a', encoding='utf-8')


def need_slurmrestd():
    """Check if slurmrestd is needed."""
    return (bc.get('DEFAULT', 'workload_scheduler') == 'Slurm'
            and not bc.get('slurm', 'use_commands'))


def init_components():
    """Initialize the components and component manager."""
    mgr = ComponentManager()

    # Slurmrestd will be started only if we're running with Slurm and
    # slurm::use_commands is not True

    @mgr.component('wf_manager', ('scheduler', 'celery'))
    def start_wfm():
        """Start the WFM."""
        fp = open_log('wf_manager')
        return launch_with_gunicorn('beeflow.wf_manager.wf_manager:create_app()',
                                    paths.wfm_socket(), stdout=fp, stderr=fp)

    tm_deps = []
    if need_slurmrestd():
        tm_deps.append('slurmrestd')

    @mgr.component('task_manager', tm_deps)
    def start_task_manager():
        """Start the TM."""
        fp = open_log('task_manager')
        return launch_with_gunicorn('beeflow.task_manager.task_manager:create_app()',
                                    paths.tm_socket(), stdout=fp, stderr=fp)

    @mgr.component('scheduler', ())
    def start_scheduler():
        """Start the scheduler."""
        fp = open_log('scheduler')
        # Using a function here because of the funny way that the scheduler's written
        return launch_with_gunicorn('beeflow.scheduler.scheduler:create_app()',
                                    paths.sched_socket(), stdout=fp, stderr=fp)

    @mgr.component('celery', ('redis',))
    def celery():
        """Start the celery task queue."""
        log = open_log('celery')
        # Setting --pool=solo to avoid preforking multiple processes
        return subprocess.Popen(['celery', '-A', 'beeflow.common.celery', 'worker', '--pool=solo'],
                                stdout=log, stderr=log)

    # Run this before daemonizing in order to avoid slow background start
    container_path = paths.redis_container()
    # If it exists, we assume that it actually has a valid container
    if not os.path.exists(container_path):
        print('Unpacking Redis image...')
        subprocess.check_call(['ch-convert', '-i', 'tar', '-o', 'dir',
                               bc.get('DEFAULT', 'redis_image'), container_path])

    @mgr.component('redis', ())
    def redis():
        """Start redis."""
        data_dir = 'data'
        os.makedirs(os.path.join(paths.redis_root(), data_dir), exist_ok=True)
        conf_name = 'redis.conf'
        container_path = paths.redis_container()
        # Dump the config
        conf_path = os.path.join(paths.redis_root(), conf_name)
        if not os.path.exists(conf_path):
            with open(conf_path, 'w', encoding='utf-8') as fp:
                # Don't listen on TCP
                print('port 0', file=fp)
                print('dir', os.path.join('/mnt', data_dir), file=fp)
                print('maxmemory 2mb', file=fp)
                print('unixsocket', os.path.join('/mnt', paths.redis_sock_fname()), file=fp)
                print('unixsocketperm 700', file=fp)
        cmd = [
            'ch-run',
            f'--bind={paths.redis_root()}:/mnt',
            container_path,
            'redis-server',
            '/mnt/redis.conf',
        ]
        log = open_log('redis')
        # Ran into a strange "Failed to configure LOCALE for invalid locale name."
        # from Redis, so setting LANG=C. This could have consequences for UTF-8
        # strings.
        env = dict(os.environ)
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        return subprocess.Popen(cmd, env=env, stdout=log, stderr=log)

    # Workflow manager and task manager need to be opened with PIPE for their stdout/stderr
    if need_slurmrestd():
        @mgr.component('slurmrestd')
        def start_slurm_restd():
            """Start BEESlurmRestD. Returns a Popen process object."""
            bee_workdir = bc.get('DEFAULT', 'bee_workdir')
            slurmrestd_log = '/'.join([bee_workdir, 'logs', 'restd.log'])
            openapi_version = bc.get('slurm', 'openapi_version')
            slurm_args = f'-s openapi/{openapi_version}'
            slurm_socket = paths.slurm_socket()
            subprocess.run(['rm', '-f', slurm_socket], check=True)
            # log.info("Attempting to open socket: {}".format(slurm_socket))
            fp = open(slurmrestd_log, 'w', encoding='utf-8') # noqa
            cmd = ['slurmrestd']
            cmd.extend(slurm_args.split())
            cmd.append(f'unix:{slurm_socket}')
            return subprocess.Popen(cmd, stdout=fp, stderr=fp)

    return mgr


MIN_CHARLIECLOUD_VERSION = (0, 34)


def version_str(version):
    """Convert a version tuple to a string."""
    return '.'.join([str(part) for part in version])


def load_check_charliecloud():
    """Load the charliecloud module if it exists and check the version."""
    if not shutil.which('ch-run'):
        lmod = os.environ.get('MODULESHOME')
        sys.path.insert(0, lmod + '/init')
        from env_modules_python import module #noqa No need to import at top
        module("load", "charliecloud")
        # Try loading the Charliecloud module then test again
        if not shutil.which('ch-run'):
            warn('Charliecloud is not loaded. Please ensure that it is accessible'
                 ' on your path.\nIf it\'s not installed on your system, please refer'
                 ' to: \n https://hpc.github.io/charliecloud/install.html')
            sys.exit(1)
    cproc = subprocess.run(['ch-run', '-V'], capture_output=True, text=True,
                           check=True)
    version = cproc.stdout if cproc.stdout else cproc.stderr
    version = version.strip()
    if 'pre' in version:
        # Pre-release charliecloud in the format <version>~pre+<git_hash>
        print(f'Found Charliecloud {version}')
        version = version.split('~')[0]
        version = tuple(int(part) for part in version.split('.'))
    # Release versions are in the format 0.<version>
    else:
        version = tuple(int(part) for part in version.split('.'))
        print(f'Found Charliecloud {version_str(version)}')
    if version < MIN_CHARLIECLOUD_VERSION:
        warn('This version of Charliecloud is too old, please upgrade to at '
             f'least version {version_str(MIN_CHARLIECLOUD_VERSION)}')
        sys.exit(1)


def check_dependencies():
    """Check for various dependencies in the environment."""
    print('Checking dependencies...')
    # Check for Charliecloud and its version
    load_check_charliecloud()
    # Check for the flux API
    if bc.get('DEFAULT', 'workload_scheduler') == 'Flux':
        try:
            import flux  # noqa needed to check whether flux api is actually installed
        except ModuleNotFoundError:
            warn('Failed to import flux Python API. Please make sure you can '
                 'use flux in your environment.')
            sys.exit(1)


class Beeflow:
    """Beeflow class for handling the main loop."""

    def __init__(self, mgr, base_components):
        """Create the Beeflow class."""
        self.mgr = mgr
        self.base_components = base_components
        self.quit = False

    def loop(self):
        """Run the main loop."""
        print(f'Running on {socket.gethostname()}')
        self.mgr.run(self.base_components)
        with cli_connection.server(paths.beeflow_socket()) as server:
            while not self.quit:
                # Handle a message from the client, if there is one
                self.handle_client(server)
                # Poll the components
                self.mgr.poll()
                time.sleep(1)
        # Kill everything, if possible
        self.mgr.kill()

    def handle_client(self, server):
        """Handle a message from the client."""
        try:
            client = server.accept()
            if client is None:
                return
            msg = client.get()
            resp = None
            if msg['type'] == 'status':
                resp = {
                    'components': self.mgr.status(),
                }
                print('Returned status info.')
            elif msg['type'] == 'quit':
                self.quit = True
                resp = 'shutting down'
                print('Shutting down.')
            client.put(resp)
        except cli_connection.BeeflowConnectionError as err:
            print(f'connection failed: {err}')


def daemonize(mgr, base_components):
    """Start beeflow as a daemon, monitoring all processes."""
    def handle_terminate(signum, stack): # noqa
        """Handle a terminate signal."""
        # Kill all subprocesses
        mgr.kill()
        sys.exit(1)

    # Now set signal handling, the log and finally daemonize
    signal_map = {
        signal.SIGINT: handle_terminate,
        signal.SIGTERM: handle_terminate,
    }
    fp = open_log('beeflow')
    with daemon.DaemonContext(signal_map=signal_map, stdout=fp, stderr=fp, stdin=fp,
                              umask=0o002):
        Beeflow(mgr, base_components).loop()


app = typer.Typer(no_args_is_help=True)


@app.command()
def start(foreground: bool = typer.Option(False, '--foreground', '-F',
          help='run in the foreground')):
    """Start all BEE components."""
    check_dependencies()
    mgr = init_components()
    beeflow_log = paths.log_fname('beeflow')
    sock_path = paths.beeflow_socket()
    if bc.get('DEFAULT', 'workload_scheduler') == 'Slurm' and not need_slurmrestd():
        warn('Not using slurmrestd. Command-line interface will be used.')
    # Note: there is a possible race condition here, however unlikely
    if os.path.exists(sock_path):
        # Try to contact for a status
        try:
            resp = cli_connection.send(sock_path, {'type': 'status'})
        except (ConnectionResetError, ConnectionRefusedError):
            resp = None
        if resp is None:
            # Must be dead, so remove the socket path
            try:
                os.remove(sock_path)
            except FileNotFoundError:
                pass
        else:
            # It's already running, so print an error and exit
            warn(f'Beeflow appears to be running. Check the beeflow log: "{beeflow_log}"')
            sys.exit(1)
    print('Starting beeflow...')
    if not foreground:
        print('Run `beeflow core status` for more information.')
    # Create the log path if it doesn't exist yet
    path = paths.log_path()
    os.makedirs(path, exist_ok=True)
    base_components = ['wf_manager', 'task_manager', 'scheduler']
    if foreground:
        try:
            Beeflow(mgr, base_components).loop()
        except KeyboardInterrupt:
            mgr.kill()
    else:
        daemonize(mgr, base_components)


@app.command()
def status():
    """Check the status of beeflow and the components."""
    resp = cli_connection.send(paths.beeflow_socket(), {'type': 'status'})
    if resp is None:
        beeflow_log = paths.log_fname('beeflow')
        warn('Cannot connect to the beeflow daemon, is it running? Check the '
             f'log at "{beeflow_log}".')
        sys.exit(1)
    print('beeflow components:')
    for comp, stat in resp['components'].items():
        print(f'{comp} ... {stat}')


@app.command()
def stop(query='yes'):
    """Stop the current running beeflow daemon."""
    if query == 'yes':
        ans = input("""
              ** Please ensure all workflows are complete before stopping beeflow. **
              ** Check the status of workflows by running 'beeflow list'.    **

              Are you sure you want to kill beeflow components? [y/n] """)
    else:
        ans = 'y'
    if ans.lower() != 'y':
        return
    resp = cli_connection.send(paths.beeflow_socket(), {'type': 'quit'})
    if resp is None:
        beeflow_log = paths.log_fname('beeflow')
        warn('Error: beeflow is not running on this system. It could be '
             'running on a different front end.\n'
             f'       Check the beeflow log: "{beeflow_log}".')
        sys.exit(1)
    # As long as it returned something, we should be good
    beeflow_log = paths.log_fname('beeflow')
    if query == "no":
        print(f'Beeflow has stopped. Check the log at "{beeflow_log}".')


def kill_active_workflows(active_states, workflow_list):
    """Kill workflows with active states."""
    db_path = wf_utils.get_db_path()
    db = connect_db(wfm_db, db_path)
    success = True
    for name, wf_id, state in workflow_list:
        if state in active_states:
            pid = db.workflows.get_gdb_pid(wf_id)
            if pid > 0:
                dep_manager.kill_gdb(pid)
            else:
                # Failure most likely caused by an Initializing workflow.
                print(f"No process for {name}, {wf_id}, {state}.")
                success = False
    return success


def archive_dir(dir_to_archive):
    """Archive directories for archive flag in reset."""
    archive_dirs = ['logs', 'container_archive', 'archives', 'workflows']
    date_str = f"{datetime.datetime.now().strftime('%Y-%m-%d_%H%M%S')}"
    backup_dir = f"{dir_to_archive}.{date_str}"
    for a_dir in archive_dirs:
        try:
            shutil.copytree(f"{dir_to_archive}/{a_dir}",
                            f"{backup_dir}/{a_dir}")
        except FileNotFoundError:
            pass
    print("Archive flag enabled.",
          "Existing logs, containers, and workflows backed up in:\n"
          f"{backup_dir}")


def handle_rm_error(err, dir_to_check, wf_list):
    """Handle IO error caused by either initializing workflows or nfs files."""
    # Check if only nfs mounts are causing the problem and ignore
    dir_list = os.listdir(dir_to_check)
    nfs_list = [x for x in dir_list if x.startswith('.nfs')]
    if dir_list and (dir_list != nfs_list):
        print(f"Unable to remove {dir_to_check} \n {err.strerror}")
        # Often initializing workflows cause a problem
        if any('Initializing' in sublist for sublist in wf_list):
            warn('Initializing workflows may have prevented removal.\n')
            print(f"Try removing {dir_to_check} manually, to complete reset.")


@app.command()
def reset(archive: bool = typer.Option(False, '--archive', '-a',
                                       help='Archive bee_workdir  before removal')):
    """Stop all components and delete the bee_workdir directory."""
    # Check workflow states; warn if there are active states.
    workflow_list = bee_client.get_wf_list()
    active_states = {'Running', 'Paused', 'Initializing', 'Waiting'}
    caution = ""
    if {item for row in workflow_list for item in row}.intersection(active_states):
        caution = """
        **************************************************************
          Caution: There are active workflows! They will be removed!
          Try 'beeflow list' to view them.
        **************************************************************
        """
    absolutely_sure = ""
    dir_to_delete = os.path.expanduser(wf_utils.get_bee_workdir())
    warn(f"\n    A reset will remove this directory: {dir_to_delete}\n")
    if archive:
        print("    Archive flag is set: logs, workflows and containers will be backed up.")
    print("""
    A reset will:
        Shutdown beeflow and all BEE components.
        Delete the bee_workdir directory which results in:
            Removing the archive of all workflows.
            Removing the archive of workflow containers
                (unless container_archive is configured elsewhere).
            Reset all databases associated with the beeflow app.
            Removing all beeflow logs.
    Beeflow configuration files from bee_cfg will not be deleted.
    """)
    warn(f"{caution}\nAre you sure you want to reset?")
    while absolutely_sure != "y" or absolutely_sure != "n":
        absolutely_sure = input("Respond with yes(y)/no(n): ")
        if absolutely_sure in ("n", "no"):
            # Exit out if the user didn't really mean to do a reset
            sys.exit()
        elif absolutely_sure in ("y", "yes"):
            # First stop all active workflow processes
            workflow_list = bee_client.get_wf_list()
            kill_active_workflows(active_states, workflow_list)
            # Stop all of the beeflow processes
            stop("quiet")
            print("Beeflow is shutting down.")
            print("Waiting for components to cleanly stop.")
            # This wait is essential. It takes a minute to shut down.
            time.sleep(5)

            # Save the bee_workdir directory if the archive option was set
            if archive:
                archive_dir(dir_to_delete)
            try:
                shutil.rmtree(dir_to_delete)
            except OSError as err:
                handle_rm_error(err, dir_to_delete, workflow_list)
            else:
                print(f"{dir_to_delete} has been removed.")
            sys.exit()
        print("Please respond with either the letter (y) or (n).")


@app.command()
def restart(foreground: bool = typer.Option(False, '--foreground', '-F',
            help='run in the foreground')):
    """Attempt to stop and restart the beeflow daemon."""
    stop()
    start(foreground)


def pull_to_tar(ref, tarball):
    """Pull a container from a registry and convert to tarball."""
    subprocess.check_call(['ch-image', 'pull', ref])
    subprocess.check_call(['ch-convert', '-i', 'ch-image', '-o', 'tar', ref, tarball])


@app.command()
def pull_deps(outdir: str = typer.Option('.', '--outdir', '-o',
                                         help='directory to store containers in')):
    """Pull required BEE containers and store in outdir."""
    load_check_charliecloud()
    neo4j_path = os.path.join(os.path.realpath(outdir), 'neo4j.tar.gz')
    pull_to_tar('neo4j:3.5.22', neo4j_path)
    redis_path = os.path.join(os.path.realpath(outdir), 'redis.tar.gz')
    pull_to_tar('redis', redis_path)
    print()
    print('The BEE dependency containers have been successfully downloaded. '
          'Please make sure to set the following options in your config:')
    print()
    print('[DEFAULT]')
    print('neo4j_image =', neo4j_path)
    print('redis_image =', redis_path)



"""Multiply Step."""

import json
import click


@click.command()
@click.argument('x', type=int)
@click.argument('y', type=int)
def multiply(x, y):
    """Multiply."""
    click.echo(json.dumps({'answer': x * y}))


if __name__ == '__main__':
    multiply(x=2, y=10)



"""Unit tests for the workflow manager."""

import tempfile
import os
import pathlib
import pytest
import jsonpickle

from test_parser import WORKFLOW_GOLD, TASKS_GOLD
from beeflow.wf_manager.wf_manager import create_app
from beeflow.wf_manager.resources import wf_utils
from beeflow.tests.mocks import MockWFI, MockGDBInterface
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.wf_interface import WorkflowInterface

from beeflow.common.db import wfm_db
from beeflow.common.db.bdb import connect_db

# We use this as the test workflow id
WF_ID = '42'


@pytest.fixture
def temp_db():
    """Pytest fixture for creating a temporary database."""
    fname = tempfile.mktemp(suffix='.db')
    db = connect_db(wfm_db, fname)
    yield db
    os.remove(fname)


@pytest.fixture
def app():
    """Create a new flask app object."""
    flask_app = create_app()
    yield flask_app


@pytest.fixture()
def client(app):
    """Return the flask test client."""
    return app.test_client


@pytest.fixture()
def teardown_workflow():
    """Just tear down for tests that use the workflow directory."""
    yield
    wf_utils.remove_wf_dir(WF_ID)


@pytest.fixture()
def setup_teardown_workflow(teardown_workflow):
    """Set up and tear down for tests that use the workflow directory."""
    wf_utils.create_workflow_dir(WF_ID)
    wf_utils.create_wf_status(WF_ID)
    yield


def _url():
    """Return URL to the workflow manager end point."""
    wfm_listen_port = bc.get('workflow_manager', 'listen_port')
    workflow_manager = 'bee_wfm/v1/jobs'
    return f'http://127.0.0.1:{wfm_listen_port}/{workflow_manager}/'


def _resource(tag=""):
    """Return the url for the workflow manager endpoint."""
    return _url() + str(tag)


class MockTask:
    """Mock task class for mocking celery."""

    @staticmethod
    def delay(*pargs, **kwargs):
        """Mock a delay call to the celery backend."""
        return None


# WFList Tests
def test_submit_workflow(client, mocker, teardown_workflow, temp_db):
    """Test submitting a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_list.init_workflow', new=MockTask)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.create_image',
                 return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.start_gdb', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.wait_gdb', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.kill_gdb', return_value=True)
    mocker.patch('beeflow.common.wf_data.generate_workflow_id', return_value='42')
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=WorkflowInterface(MockGDBInterface()))
    mocker.patch('subprocess.run', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_db_path', temp_db.db_file)
    mocker.patch('beeflow.wf_manager.resources.wf_list.db_path', temp_db.db_file)
    script_path = pathlib.Path(__file__).parent.resolve()
    tarball = script_path / 'clamr-wf.tgz'
    with open(tarball, 'rb') as tarball_contents:
        resp = client().post('/bee_wfm/v1/jobs/', data={
            'wf_name': 'clamr'.encode(),
            'wf_filename': tarball,
            'workdir': '.',
            'workflow': jsonpickle.encode(WORKFLOW_GOLD),
            'tasks': jsonpickle.encode(TASKS_GOLD, warn=True),
            'workflow_archive': tarball_contents,
            'no_start': False
        })

    # Remove task added during the test
    assert resp.json['msg'] == 'Workflow uploaded'


def test_reexecute_workflow(client, mocker, teardown_workflow, temp_db):
    """Test reexecuting a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_list.init_workflow', new=MockTask)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.create_image',
                 return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.start_gdb', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.wait_gdb', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.kill_gdb', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_db_path', temp_db.db_file)
    mocker.patch('beeflow.wf_manager.resources.wf_list.db_path', temp_db.db_file)
    mocker.patch('beeflow.common.wf_data.generate_workflow_id', return_value='42')
    mocker.patch('subprocess.run', return_value=True)

    script_path = pathlib.Path(__file__).parent.resolve()
    tarball = script_path / '42.tgz'
    with open(tarball, 'rb') as tarball_contents:
        resp = client().put('/bee_wfm/v1/jobs/', data={
                            'wf_filename': tarball,
                            'wf_name': 'clamr',
                            'workflow_archive': tarball_contents,
                            'workdir': '.'
                            })

        assert resp.json['msg'] == 'Workflow uploaded'


class MockDBWorkflowHandle:
    """Mock DB workflow handle."""

    def update_workflow_state(self, *pargs, **kwargs):
        """Mock update a workflow."""


class MockDB:
    """Mock DB for workflow manager."""

    @property
    def workflows(self):
        """Return a workflow handle."""
        return MockDBWorkflowHandle()


def mock_connect_db(*pargs, **kwargs):
    """Mock a DB connection."""
    return MockDB()


# WFActions Tests
def test_start_workflow(client, mocker, temp_db):
    """Test starting a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.connect_db', new=mock_connect_db)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.wf_manager.resources.wf_utils.submit_tasks_tm', return_value=None)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.submit_tasks_scheduler', return_value=None)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.update_wf_status', return_value=None)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_db_path', new=lambda: temp_db.db_file)
    mocker.patch('beeflow.wf_manager.resources.wf_actions.db_path', temp_db.db_file)
    resp = client().post(f'/bee_wfm/v1/jobs/{WF_ID}')
    assert resp.status_code == 200


def test_workflow_status(client, mocker, setup_teardown_workflow, temp_db):
    """Test getting workflow status."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_db_path', temp_db.db_file)
    mocker.patch('beeflow.wf_manager.resources.wf_actions.db_path', temp_db.db_file)
    wf_name = 'wf'
    bolt_port = 3030
    http_port = 3333
    https_port = 3455

    temp_db.workflows.init_workflow(WF_ID, wf_name, 'dir', bolt_port, http_port, https_port)
    temp_db.workflows.add_task(123, WF_ID, 'task', "WAITING")
    temp_db.workflows.add_task(124, WF_ID, 'task', "RUNNING")

    resp = client().get(f'/bee_wfm/v1/jobs/{WF_ID}')
    tasks_status = resp.json['tasks_status']
    assert tasks_status[0][2] == 'RUNNING' or tasks_status[1][2] == 'RUNNING'


def test_cancel_workflow(client, mocker, setup_teardown_workflow, temp_db):
    """Test cancelling a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_db_path', temp_db.db_file)
    mocker.patch('beeflow.wf_manager.resources.wf_actions.db_path', temp_db.db_file)

    wf_name = 'wf'
    wf_status = 'Pending'
    bolt_port = 3030
    gdb_pid = 12345

    temp_db.workflows.init_workflow(WF_ID, wf_name, wf_status, 'dir', bolt_port, gdb_pid)
    temp_db.workflows.add_task(123, WF_ID, 'task', "WAITING")
    temp_db.workflows.add_task(124, WF_ID, 'task', "RUNNING")
    mocker.patch('beeflow.wf_manager.resources.wf_actions.dep_manager.kill_gdb', return_value=None)

    request = {'wf_id': WF_ID, 'option': 'cancel'}
    resp = client().delete(f'/bee_wfm/v1/jobs/{WF_ID}', json=request)
    assert resp.json['status'] == 'Cancelled'
    assert resp.status_code == 202


def test_remove_workflow(client, mocker, setup_teardown_workflow, temp_db):
    """Test removing a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_db_path', temp_db.db_file)
    mocker.patch('beeflow.wf_manager.resources.wf_actions.db_path', temp_db.db_file)

    wf_name = 'wf'
    wf_status = 'Archived'
    bolt_port = 3030
    gdb_pid = 12345

    temp_db.workflows.init_workflow(WF_ID, wf_name, wf_status, 'dir', bolt_port, gdb_pid)
    temp_db.workflows.add_task(123, WF_ID, 'task', "WAITING")
    temp_db.workflows.add_task(124, WF_ID, 'task', "RUNNING")
    mocker.patch('beeflow.wf_manager.resources.wf_actions.dep_manager.kill_gdb', return_value=None)

    request = {'wf_id': WF_ID, 'option': 'remove'}
    resp = client().delete(f'/bee_wfm/v1/jobs/{WF_ID}', json=request)
    assert resp.json['status'] == 'Removed'
    assert resp.status_code == 202


def test_pause_workflow(client, mocker, setup_teardown_workflow, temp_db):
    """Test pausing a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.tests.mocks.MockWFI.get_workflow_state',
                 return_value='RUNNING')
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_db_path', temp_db.db_file)
    mocker.patch('beeflow.wf_manager.resources.wf_actions.db_path', temp_db.db_file)

    wf_utils.update_wf_status(WF_ID, 'Running')
    request = {'option': 'pause'}
    resp = client().patch(f'/bee_wfm/v1/jobs/{WF_ID}', json=request)
    assert resp.json['status'] == 'Workflow Paused'
    assert resp.status_code == 200


def test_resume_workflow(client, mocker, setup_teardown_workflow, temp_db):
    """Test resuming a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.tests.mocks.MockWFI.get_workflow_state',
                 return_value='PAUSED')
    mocker.patch('beeflow.wf_manager.resources.wf_utils.submit_tasks_tm', return_value=None)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.submit_tasks_scheduler', return_value=None)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.update_wf_status', return_value=None)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_db_path', temp_db.db_file)
    mocker.patch('beeflow.wf_manager.resources.wf_actions.db_path', temp_db.db_file)

    wf_utils.update_wf_status(WF_ID, 'Paused')
    request = {'option': 'resume'}
    resp = client().patch(f'/bee_wfm/v1/jobs/{WF_ID}', json=request)
    assert resp.json['status'] == 'Workflow Resumed'
    assert resp.status_code == 200
# pylama:ignore=W0621,W0613



"""Linear Regression."""
import json
import pickle
import click
import numpy as np
from sklearn.linear_model import LinearRegression


@click.command()
@click.argument('x1', type=int)   # Enter number of iteration
def reg(x1):
    """Linear Regression."""
    X = pickle.load(open('/home/bee/cwl2/MyX.p', 'rb'))
    Y = pickle.load(open('/home/bee/cwl2/MyY.p', 'rb'))

    X = X.values

    print("My pickle X is", X)
    Y = Y.values
    print("My pickle Y is", Y)
    m = []
    y_in = []
    for i in range(x1):
        reg = LinearRegression().fit(X, Y)

        pickle.dump(reg, open('mymodel.p', 'wb'))
        m.append(reg.coef_)
        y_in.append(reg.intercept_)
        # m, y_in = reg.coef_, reg.intercept_

    average_slope = np.mean(m)
    average_y_intercept = np.mean(y_in)
    m_list = average_slope.tolist()
    m_json = json.dumps(m_list)
    y_in_list = average_y_intercept.tolist()
    y_in_json = json.dumps(y_in_list)

    print('Learning regression line parameters.')
    print(average_slope, average_y_intercept)  # model parameters i.e. slope and y-intercept


if __name__ == '__main__':
    reg(x1=1)
# Ignores preserving code for now
# pylama:ignore=C0103,R1732,W0612,W0621



"""Task Manager utility functions."""
import os
from pathlib import Path
import re
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.db import tm_db
from beeflow.common.db import bdb
from beeflow.common import worker
from beeflow.common import paths
from beeflow.common.connection import Connection
from beeflow.common.worker_interface import WorkerInterface


def db_path():
    """Return the TM database path."""
    bee_workdir = bc.get('DEFAULT', 'bee_workdir')
    return os.path.join(bee_workdir, 'tm.db')


def connect_db():
    """Connect to the TM database."""
    return bdb.connect_db(tm_db, db_path())


def worker_interface():
    """Load the worker interface."""
    wls = bc.get('DEFAULT', 'workload_scheduler')
    worker_class = worker.find_worker(wls)
    if worker_class is None:
        raise RuntimeError(f'Workload scheduler {wls}, not supported.\n'
                           + f'Please check {bc.userconfig_path()} and restart TaskManager.')
    # Get the parameters for the worker classes
    worker_kwargs = {
        'bee_workdir': bc.get('DEFAULT', 'bee_workdir'),
        'container_runtime': bc.get('task_manager', 'container_runtime'),
        # extra options to be passed to the runner (i.e. srun [RUNNER_OPTS] ... for Slurm)
        'runner_opts': bc.get('task_manager', 'runner_opts'),
    }
    # Job defaults
    for default_key in ['default_account', 'default_time_limit', 'default_partition']:
        worker_kwargs[default_key] = bc.get('job', default_key)
    # Special slurm arguments
    if wls == 'Slurm':
        worker_kwargs['use_commands'] = bc.get('slurm', 'use_commands')
        worker_kwargs['slurm_socket'] = paths.slurm_socket()
        worker_kwargs['openapi_version'] = bc.get('slurm', 'openapi_version')
    return WorkerInterface(worker_class, **worker_kwargs)


def wfm_url():
    """Return the url to the WFM."""
    # Saving this for whenever we need to run jobs across different machines
    # workflow_manager = 'bee_wfm/v1/jobs/'
    # #wfm_listen_port = bc.get('workflow_manager', 'listen_port')
    # wfm_listen_port = wf_db.get_wfm_port()
    # return f'http://127.0.0.1:{wfm_listen_port}/{workflow_manager}'
    return 'bee_wfm/v1/jobs/'


def wfm_resource_url(tag=""):
    """Get the full resource URL for access to the WFM."""
    return wfm_url() + str(tag)


def wfm_conn():
    """Get a new connection to the WFM."""
    return Connection(paths.wfm_socket())


class CheckpointRestartError(Exception):
    """Exception to be thrown on checkpoint-restart failure."""


def get_restart_file(task_checkpoint, task_workdir):
    """Find latest checkpoint file."""
    if 'file_regex' not in task_checkpoint:
        raise CheckpointRestartError('file_regex is required for checkpointing')
    if 'file_path' not in task_checkpoint:
        raise CheckpointRestartError('file_path is required for checkpointing')
    file_regex = task_checkpoint['file_regex']
    file_path = Path(task_workdir, task_checkpoint['file_path'])
    regex = re.compile(file_regex)
    try:
        checkpoint_files = [
            Path(file_path, fname) for fname in os.listdir(file_path)
            if regex.match(fname)
        ]
    except FileNotFoundError:
        raise CheckpointRestartError(
            f'Checkpoint file_path ("{file_path}") not found'
        ) from None
    checkpoint_files.sort(key=os.path.getmtime)
    try:
        checkpoint_file = checkpoint_files[-1]
        return str(checkpoint_file)
    except IndexError:
        raise CheckpointRestartError('Missing checkpoint file for task') from None



"""Flux worker interface."""

import os
from beeflow.common import log as bee_logging
from beeflow.common.worker.worker import Worker

log = bee_logging.setup(__name__)

# Map from flux states to BEE statuses
BEE_STATES = {
    'NEW': 'PENDING',
    'DEPEND': 'PENDING',
    'PRIORITY': 'PENDING',
    'SCHED': 'PENDING',
    'RUN': 'RUNNING',
    'CLEANUP': 'COMPLETING',
    'INACTIVE': 'COMPLETED',
    'COMPLETED': 'COMPLETED',
    'FAILED': 'FAILED',
    'CANCELED': 'CANCELED',
    'TIMEOUT': 'TIMEOUT',
}


class FluxWorker(Worker):
    """Flux worker code."""

    def __init__(self, **kwargs):
        """Initialize the flux worker object."""
        super().__init__(**kwargs)
        # Only try to import the Flux API if we need it
        import flux   # noqa this is necessary since flux may not be installed
        from flux import job  # noqa
        self.flux = flux
        self.job = job

    def build_text(self, task):
        """Build text for task script."""
        # TODO: Not used for the Flux worker

    def build_jobspec(self, task):
        """Build the job spec for a task."""
        # TODO: This has a lot of code in common with the other worker's build_text
        crt_res = self.crt.run_text(task)
        script = [
            '#!/bin/bash',
            'set -e',
            crt_res.env_code,
        ]

        # TODO: Should this entire model, saving stdout and stderr to files, be
        # redone for Flux? It seems to provide some sort of KVS for storing
        # output but I don't quite understand it.

        # Get resource requirements
        nodes = task.get_requirement('beeflow:MPIRequirement', 'nodes', default=1)
        # TODO: 'ntasks' may not mean the same thing as with Slurm
        ntasks = task.get_requirement('beeflow:MPIRequirement', 'ntasks', default=nodes)
        # TODO: What to do with the MPI version?
        # mpi_version = task.get_requirement('beeflow:MPIRequirement', 'mpiVersion',
        #                                    default='pmi2')

        for cmd in crt_res.pre_commands:
            if cmd.type == 'one-per-node':
                cmd_args = ['flux', 'run', '-N', str(nodes), '-n', str(nodes), ' '.join(cmd.args)]
            else:
                cmd_args = ['flux', 'run', ' '.join(cmd.args)]
            script.append(' '.join(cmd_args))

        # Set up the main command
        args = ['flux', 'run', '-N', str(nodes), '-n', str(ntasks)]
        if task.stdout is not None:
            args.extend(['--output', task.stdout])
        if task.stderr is not None:
            args.extend(['--error', task.stderr])
        args.extend(crt_res.main_command.args)
        log.info(args)
        script.append(' '.join(args))

        for cmd in crt_res.post_commands:
            if cmd.type == 'one-per-node':
                cmd_args = ['flux', 'run', '-N', str(nodes), '-n', str(nodes), ' '.join(cmd.args)]
            else:
                cmd_args = ['flux', 'run', ' '.join(cmd.args)]
            script.append(' '.join(cmd_args))

        script = '\n'.join(script)
        jobspec = self.job.JobspecV1.from_batch_command(script, task.name,
                                                        num_slots=ntasks,
                                                        num_nodes=nodes)
        task_save_path = self.task_save_path(task)
        jobspec.stdout = f'{task_save_path}/{task.name}-{task.id}.out'
        jobspec.stderr = f'{task_save_path}/{task.name}-{task.id}.err'
        jobspec.environment = dict(os.environ)
        # Save the script for later reference
        with open(f'{task_save_path}/{task.name}-{task.id}.sh', 'w', encoding='utf-8') as fp:
            fp.write(script)
        return jobspec

    def submit_task(self, task):
        """Worker submits task; returns job_id, job_state."""
        log.info(f'Submitting task: {task.name}')
        jobspec = self.build_jobspec(task)
        flux = self.flux.Flux()
        job_id = self.job.submit(flux, jobspec)
        return job_id, self.query_task(job_id)

    def cancel_task(self, job_id):
        """Cancel task with job_id; returns job_state."""
        log.info(f'Cancelling task with ID: {job_id}')
        flux = self.flux.Flux()
        self.job.cancel(flux, job_id)
        return 'CANCELED'

    def query_task(self, job_id):
        """Query job state for the task."""
        # TODO: How does Flux handle TIMEOUT/TIMELIMIT? They don't seem to have
        # a state for this
        log.info(f'Querying task with job_id: {job_id}')
        flux = self.flux.Flux()
        info = self.job.get_job(flux, job_id)
        log.info(info)

        # TODO: May need to check for return codes other than 0 if
        # specified by the task (although I'm not sure how we can keep
        # track of this with job ID alone)

        # Note: using 'status' here instead of 'state'
        return BEE_STATES[info['status']]
# Ignoring W0511: TODO's are needed here to indicate parts of the code that may
#                 need more work or thought
# pylama:ignore=W0511



#!/usr/bin/env python3
"""bee-client.

This script provides an client interface to the user to manage workflows.
Capablities include submitting, starting, listing, pausing and cancelling workflows.
"""
import os
import sys
import logging
import inspect
import pathlib
import shutil
import subprocess
import tempfile
import textwrap
import time
import importlib.metadata
import jsonpickle
import requests
import typer

from beeflow.common import config_driver
from beeflow.common.cli import NaturalOrderGroup
from beeflow.common.connection import Connection
from beeflow.common import paths
from beeflow.common.parser import CwlParser
from beeflow.common.wf_data import generate_workflow_id
from beeflow.client import core
from beeflow.wf_manager.resources import wf_utils

# Length of a shortened workflow ID
short_id_len = 6 #noqa: Not a constant

# Maximum length of a workflow ID
MAX_ID_LEN = 32

# Global used to indicate whether this instance is interactive or not
_INTERACTIVE = False


logging.basicConfig(level=logging.WARNING)
WORKFLOW_MANAGER = 'bee_wfm/v1/jobs/'


class ClientError(Exception):
    """Client error class."""

    def __init__(self, *args):
        """Error constructor."""
        self.args = args


def error_exit(msg, include_caller=True):
    """Print a message and exit or raise an error with that message."""
    if include_caller:
        caller_func = str.capitalize(inspect.stack()[1].function)
        msg = caller_func + ': ' + msg
    if _INTERACTIVE:
        typer.secho(msg, fg=typer.colors.RED, file=sys.stderr)
        sys.exit(1)
    else:
        # Raise an error so that client libraries can handle it
        raise ClientError(msg) from None


def error_handler(resp):  # noqa (this is an error handler, it doesn't need to return an expression)
    """Handle a 500 error in a response."""
    if resp.status_code != 500:
        return resp
    data = resp.json()
    if 'error' not in data:
        return resp
    error = data['error']
    error_file = f'bee-error-{int(time.time())}.log'
    # Save the args and exception info to the file
    with open(error_file, 'w', encoding='utf-8') as fp:
        fp.write(f'args: {" ".join(sys.argv)}\n')
        fp.write(error)
    msg = ('An error occurred with the WFM. Please save your workflow and the '
           f'error log "{error_file}". If possible, please report this to the '
           'BEE development team.')
    msg = textwrap.fill(msg)
    error_exit(msg, include_caller=False)


def _wfm_conn():
    """Return a connection to the WFM."""
    return Connection(paths.wfm_socket(),
                      error_handler=error_handler)


def _url():
    #    """Returns URL to the workflow manager end point"""
    #    wfm_listen_port = wf_db.get_wfm_port()
    #    workflow_manager = 'bee_wfm/v1/jobs'
    #    return f'http://127.0.0.1:{wfm_listen_port}/{workflow_manager}/'
    """Return URL to the workflow manager end point."""
    return WORKFLOW_MANAGER


def _short_id(wf_id):
    return wf_id[:short_id_len]


def _resource(tag=""):
    return _url() + str(tag)


def get_wf_list():
    """Get the list of all workflows."""
    try:
        conn = _wfm_conn()
        resp = conn.get(_url(), timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('WF Manager did not return workflow list')

    logging.info('List Jobs:  {resp.text}')
    return jsonpickle.decode(resp.json()['workflow_list'])


def check_short_id_collision():
    """Check short workflow IDs for colliions; increase short ID length if detected."""
    global short_id_len  #noqa: Not a constant
    workflow_list = get_wf_list()
    if workflow_list:
        while short_id_len < MAX_ID_LEN:
            id_list = [_short_id(job[1]) for job in workflow_list]
            id_list_set = set(id_list)
            # Collision if set shorter than list
            if len(id_list_set) < len(id_list):
                short_id_len += 1
            else:
                break
        else:
            raise RuntimeError("collision detected between two full workflow IDs")
    else:
        print("There are currently no jobs.")


def match_short_id(wf_id):
    """Match user-provided short workflow ID to full workflow IDs."""
    matched_ids = []
    workflow_list = get_wf_list()
    if workflow_list:
        for job in workflow_list:
            if job[1].startswith(wf_id):
                matched_ids.append(job[1])
        if len(matched_ids) > 1:
            logging.info(f"user-provided workflow ID {wf_id} matched multiple stored workflow IDs")
            error_exit("provided workflow ID ambiguous")
        elif not matched_ids:
            logging.info(f"user-provided workflow ID {wf_id} did not match any"
                         "stored workflow ID")
            error_exit("Provided workflow ID does not match any submitted "
                       "workflows")
        else:
            logging.info(f"user-provided workflow ID {wf_id} matched stored"
                         "workflow ID {matched_ids[0]}")
            long_wf_id = matched_ids[0]
            return long_wf_id
    else:
        sys.exit("There are currently no workflows.")

    return None


def get_wf_status(wf_id):
    """Get workflow status."""
    try:
        conn = _wfm_conn()
        resp = conn.get(_resource(wf_id), timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('Could not successfully query workflow manager')

    return resp.json()['wf_status']


app = typer.Typer(no_args_is_help=True, add_completion=False, cls=NaturalOrderGroup)
app.add_typer(core.app, name='core')
app.add_typer(config_driver.app, name='config')


@app.command()
def submit(wf_name: str = typer.Argument(..., help='the workflow name'),  # pylint:disable=R0915
           wf_path: pathlib.Path = typer.Argument(..., help='path to the workflow .tgz or dir'),
           main_cwl: str = typer.Argument(...,
           help='filename of main CWL (if using CWL tarball), '
           + 'path of main CWL (if using CWL directory)'),
           yaml: str = typer.Argument(...,
           help='filename of yaml file (if using CWL tarball), '
           + 'path of yaml file (if using CWL directory)'),
           workdir: pathlib.Path = typer.Argument(...,
           help='working directory for workflow containing input + output files',),
           no_start: bool = typer.Option(False, '--no-start', '-n',
                                         help='do not start the workflow')):
    """Submit a new workflow."""
    def is_parent(parent, path):
        """Return true if the path is a child of the other path."""
        parent = os.path.abspath(parent)
        path = os.path.abspath(path)
        return os.path.commonpath([parent]) == os.path.commonpath([parent, path])

    tarball_path = ""
    if os.path.exists(wf_path):
        # Check to see if the wf_path is a tarball or a directory. Package if directory
        if os.path.isdir(wf_path):
            print("Detected directory instead of packaged workflow. Packaging Directory...")
            main_cwl_path = pathlib.Path(main_cwl).resolve()
            yaml_path = pathlib.Path(yaml).resolve()

            if not main_cwl_path.exists():
                error_exit(f'Main CWL file {main_cwl} does not exist')
            if not yaml_path.exists():
                error_exit(f'YAML file {yaml} does not exist')

            # Packaging in temp dir, after copying alternate cwl_main or yaml file
            cwl_indir = is_parent(wf_path, main_cwl_path)
            yaml_indir = is_parent(wf_path, yaml_path)
            tempdir_path = pathlib.Path(tempfile.mkdtemp())
            if cwl_indir and yaml_indir:
                package_path = package(wf_path, tempdir_path)
            else:
                tempdir_wf_path = pathlib.Path(tempdir_path / wf_path.name)
                shutil.copytree(wf_path, tempdir_wf_path, dirs_exist_ok=False)
                if not cwl_indir:
                    shutil.copy2(main_cwl, tempdir_wf_path)
                if not yaml_indir:
                    shutil.copy2(yaml, tempdir_wf_path)
                package_path = package(tempdir_wf_path, tempdir_path)
        else:
            package_path = wf_path
        # Untar and parse workflow
        untar_path = pathlib.Path(tempfile.mkdtemp())
        untar_wf_path = unpackage(package_path, untar_path)
        main_cwl_path = untar_wf_path / pathlib.Path(main_cwl).name
        yaml_path = untar_wf_path / pathlib.Path(yaml).name
        parser = CwlParser()
        workflow_id = generate_workflow_id()
        workflow, tasks = parser.parse_workflow(workflow_id, str(main_cwl_path),
                                                job=str(yaml_path))
        tasks = [jsonpickle.encode(task) for task in tasks]

        wf_tarball = open(package_path, 'rb')
        shutil.rmtree(untar_path)
        if os.path.isdir(wf_path):
            shutil.rmtree(tempdir_path)
    else:
        error_exit(f'Workflow tarball {wf_path} cannot be found')

    # Make sure the workdir is an absolute path and exists
    workdir = os.path.expanduser(workdir)
    workdir = os.path.abspath(workdir)
    if not os.path.exists(workdir):
        error_exit(f"Workflow working directory \"{workdir}\" doesn't exist")

    # TODO: Can all of this information be sent as a file?
    data = {
        'wf_name': wf_name.encode(),
        'wf_filename': os.path.basename(wf_path).encode(),
        'workdir': workdir,
        'workflow': jsonpickle.encode(workflow),
        'tasks': jsonpickle.encode(tasks, warn=True),
        'no_start': no_start,
    }
    files = {
        'workflow_archive': wf_tarball
    }
    try:
        conn = _wfm_conn()
        resp = conn.post(_url(), data=data, files=files, timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.created:  # pylint: disable=no-member
        if resp.status_code == 400:
            data = resp.json()
            error_exit(data['msg'])
        error_exit(f"Submit for {wf_name} failed. Please check the WF Manager.")

    check_short_id_collision()
    if 'wf_id' not in resp.json():
        error_exit("wf_id not in WFM response")
    wf_id = resp.json()['wf_id']
    typer.secho("Workflow submitted! Your workflow id is "
                f"{_short_id(wf_id)}.", fg=typer.colors.GREEN)
    logging.info('Sumit workflow:  {resp.text}')

    # Cleanup code
    if tarball_path:
        os.remove(tarball_path)

    # Store provided arguments in text file for future reference
    wf_dir = wf_utils.get_workflow_dir(wf_id)
    sub_wf_dir = wf_dir + "/submit_command_args.txt"

    f_name = open(sub_wf_dir, "w", encoding="utf-8")
    f_name.write(f"wf_name: {wf_name}\n")
    f_name.write(f"wf_path: {wf_path}\n")
    f_name.write(f"main_cwl: {main_cwl}\n")
    f_name.write(f"yaml: {yaml}\n")
    f_name.write(f"workdir: {workdir}\n")
    f_name.write(f"wf_id: {wf_id}")
    f_name.close()

    return wf_id


@app.command()
def start(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Start a workflow with a workflow ID."""
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.post(_resource(long_wf_id), timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code == 400:
        error_exit("Could not start workflow. It may have already been started "
                   "and ran to completion (or failure).")
    elif resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit(f"Starting {long_wf_id} failed."
                   f" Returned {resp.status_code}")

    typer.echo(f"{resp.json()['msg']}")
    logging.info('Started  {resp.text}')


@app.command()
def package(wf_path: pathlib.Path = typer.Argument(...,
            help='Path to the workflow package directory'),
            package_dest: pathlib.Path = typer.Argument(...,
            help='Path for where the packaged workflow should be saved')
            ):
    """Package a workflow into a tarball."""
    if not os.path.isdir(wf_path):
        error_exit(f"{wf_path} is not a valid directory.")

    # Get the filename
    wf_dir = wf_path.name
    tarball = wf_path.name + '.tgz'
    # Get the parent directory and resolve it in case the path is relative
    parent_dir = wf_path.resolve().parent
    # Just use tar with subprocess. Python's tar library is not performant.
    return_code = subprocess.run(['tar', '-C', parent_dir, '-czf', tarball, wf_dir],
                                 check=True).returncode
    package_path = package_dest.resolve()/tarball  # noqa: Not an arithmetic operation

    # Get the curent working directory
    cwd = pathlib.Path().absolute()
    if package_dest != cwd:
        # Move the tarball if the directory it's wanted in is not in the current working directory
        tarball_path = cwd/tarball  # noqa: Not an arithmetic operation
        shutil.move(tarball_path, package_path)

    if return_code != 0:
        error_exit("Package failed")
    else:
        print(f"Package {tarball} created successfully")

    return package_path


@app.command()
def remove(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Remove cancelled, paused, or archived workflow with a workflow ID."""
    long_wf_id = wf_id

    wf_status = get_wf_status(wf_id)
    print(f"Workflow Status is {wf_status}")
    if wf_status in ('Cancelled', 'Archived', 'Paused'):
        verify = f"All stored information for workflow {_short_id(wf_id)} will be removed."
        verify += "\nContinue to remove? yes(y)/no(n): """
        response = input(verify)
        if response in ("n", "no"):
            sys.exit("Workflow not removed.")
        elif response in ("y", "yes"):
            try:
                conn = _wfm_conn()
                resp = conn.delete(_resource(long_wf_id), json={'option': 'remove'}, timeout=60)
            except requests.exceptions.ConnectionError:
                error_exit('Could not reach WF Manager.')
            if resp.status_code != requests.codes.accepted:  # pylint: disable=no-member
                error_exit('WF Manager could not remove workflow.')
            typer.secho("Workflow removed!", fg=typer.colors.GREEN)
            logging.info(f'Remove workflow: {resp.text}')
    else:
        print(f"{_short_id(wf_id)} may still be running.")
        print("The workflow must be cancelled before attempting removal.")

    sys.exit()


def unpackage(package_path, dest_path):
    """Unpackage a workflow tarball for parsing."""
    package_str = str(package_path)
    package_path = package_path.resolve()

    if not package_str.endswith('.tgz'):
        # No cleanup, maybe we should rm dest_path?
        error_exit("Invalid package name, please use the beeflow package command")
    wf_dir = pathlib.Path(package_path).stem

    return_code = subprocess.run(['tar', '-C', dest_path, '-xf', package_path],
                                 check=True).returncode
    if return_code != 0:
        # No cleanup, maybe we should rm dest_path?
        error_exit("Unpackage failed")
    else:
        print(f"Package {package_str} unpackaged successfully")
    return pathlib.Path(dest_path / wf_dir)


@app.command('list')
def list_workflows():
    """List all workflows."""
    workflow_list = get_wf_list()
    if workflow_list:
        typer.secho("Name\tID\tStatus", fg=typer.colors.GREEN)

        for name, wf_id, status in workflow_list:
            typer.echo(f"{name}\t{_short_id(wf_id)}\t{status}")
    else:
        typer.echo("There are currently no workflows.")

    logging.info('List workflows:  {resp.text}')


@app.command()
def query(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Get the status of a workflow."""
    # wf_id is a tuple with the short version and long version
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.get(_resource(long_wf_id), timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('Could not successfully query workflow manager')

    tasks_status = resp.json()['tasks_status']
    wf_status = resp.json()['wf_status']
    typer.echo(wf_status)
    for _task_id, task_name, task_state in tasks_status:
        typer.echo(f'{task_name}--{task_state}')

    logging.info('Query workflow:  {resp.text}')
    return wf_status, tasks_status


@app.command()
def metadata(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Get metadata about a given workflow."""
    try:
        conn = _wfm_conn()
        resp = conn.get(_resource(wf_id) + '/metadata', timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.okay:  # noqa (pylama doesn't know about the okay member)
        error_exit('Could not successfully query workflow manager')

    # Print and or return the metadata
    data = resp.json()
    for key, value in data.items():
        typer.echo(f'{key} = {value}')
    return data


@app.command()
def pause(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Pause a workflow (Running tasks will finish)."""
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.patch(_resource(long_wf_id), json={'option': 'pause'},
                          timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')
    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('WF Manager could not pause workflow.')
    logging.info('Pause workflow:  {resp.text}')


@app.command()
def resume(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Resume a paused workflow."""
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.patch(_resource(long_wf_id),
                          json={'wf_id': long_wf_id, 'option': 'resume'},
                          timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')
    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('WF Manager could not resume workflow.')
    logging.info('Resume workflow:  {resp.text}')


@app.command()
def cancel(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Cancel a paused or running workflow."""
    long_wf_id = wf_id
    wf_status = get_wf_status(wf_id)
    if wf_status in ('Running', 'Paused'):
        try:
            conn = _wfm_conn()
            resp = conn.delete(_resource(long_wf_id), json={'option': 'cancel'}, timeout=60)

        except requests.exceptions.ConnectionError:
            error_exit('Could not reach WF Manager.')
        if resp.status_code != requests.codes.accepted:  # pylint: disable=no-member
            error_exit('WF Manager could not cancel workflow.')
        typer.secho("Workflow cancelled!", fg=typer.colors.GREEN)
        logging.info(f'Cancel workflow: {resp.text}')
    elif wf_status == "Intializing":
        print(f"Workflow is {wf_status}, try cancel later.")
    else:
        print(f"Workflow is {wf_status} cannot cancel.")


@app.command()
def copy(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Copy an archived workflow."""
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.patch(_url(), files={'wf_id': long_wf_id}, timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')
    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('WF Manager could not copy workflow.')
    archive_file = jsonpickle.decode(resp.json()['archive_file'])
    archive_filename = resp.json()['archive_filename']
    logging.info(f'Copy workflow: {resp.text}')
    return archive_file, archive_filename


@app.command()
def reexecute(wf_name: str = typer.Argument(..., help='The workflow name'),
              wf_path: pathlib.Path = typer.Argument(..., help='Path to the workflow archive'),
              workdir: pathlib.Path = typer.Argument(
                  ...,
                  help='working directory for workflow containing input + output files')
              ):
    """Reexecute an archived workflow."""
    if os.path.exists(wf_path):
        wf_tarball = open(wf_path, 'rb')
    else:
        error_exit(f'Workflow tarball {wf_path} cannot be found')

    data = {
        'wf_filename': os.path.basename(wf_path).encode(),
        'wf_name': wf_name.encode(),
        'workdir': workdir
    }
    try:
        conn = _wfm_conn()
        resp = conn.put(_url(), data=data,
                        files={'workflow_archive': wf_tarball}, timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.created: #noqa: member does exist
        error_exit(f"Reexecute for {wf_name} failed. Please check the WF Manager.")

    wf_id = resp.json()['wf_id']
    typer.secho("Workflow submitted! Your workflow id is "
                f"{_short_id(wf_id)}.", fg=typer.colors.GREEN)
    logging.info(f'ReExecute Workflow: {resp.text}')
    return wf_id


@app.callback(invoke_without_command=True)
def version_callback(version: bool = False):
    """Beeflow."""
    # Print out the current version of the app, and then exit
    # Note above docstring gets used in the help menu
    if version:
        version = importlib.metadata.version("hpc-beeflow")
        print(version)


def main():
    """Execute bee_client."""
    global _INTERACTIVE
    _INTERACTIVE = True
    app()


if __name__ == "__main__":
    app()

# Ignore W0511: This allows us to have TODOs in the code
# Ignore R1732: Significant code restructuring required to fix
# pylama:ignore=W0511,R1732



"""Add Step."""
import json
import click


@click.command()
@click.argument('x', type=int)
@click.argument('y', type=int)
def add(x, y):
    """Add."""
    click.echo(json.dumps({'answer': x + y}))


if __name__ == '__main__':
    add(x=1, y=2)



"""Scheduler database code."""

import jsonpickle
from beeflow.common.db import bdb


class Resources:
    """Resources wrapper class."""

    def __init__(self, db_file):
        """Create the resources class."""
        self.db_file = db_file

    def __iter__(self):
        """Create an iterator over the resources."""
        stmt = 'SELECT id, resource FROM resources'
        result = bdb.getall(self.db_file, stmt)
        for rslt in result:
            resource = jsonpickle.decode(rslt[1])
            yield resource

    def extend(self, resources):
        """Add a list of new resources."""
        for resource in resources:
            data = jsonpickle.encode(resource)
            stmt = 'INSERT INTO resources (resource) VALUES (?)'
            bdb.run(self.db_file, stmt, [data])

    def clear(self):
        """Remove all resources."""
        clear_stmt = 'DELETE FROM resources'
        bdb.run(self.db_file, clear_stmt)


class SchedDB:
    """Scheduler database."""

    def __init__(self, db_file):
        """Construct a new scheduler database connection."""
        self.db_file = db_file
        self._init_tables()

    def _init_tables(self):
        """Initialze the scheduler tables if they don't exist."""
        resource_stmt = """CREATE TABLE IF NOT EXISTS resources(
                        id INTEGER PRIMARY KEY ASC,
                        resource TEXT);"""
        bdb.create_table(self.db_file, resource_stmt)

    @property
    def resources(self):
        """Get resources from the database."""
        return Resources(self.db_file)


def open_db(db_file):
    """Open and return a new database."""
    return SchedDB(db_file)



"""Abstract base class for worker, the workload manager."""

from abc import ABC, abstractmethod
import os
from beeflow.common import log as bee_logging
from beeflow.common.crt_interface import ContainerRuntimeInterface


log = bee_logging.setup(__name__)


# Import all implemented container runtime drivers now
# No error if they don't exist
try:
    from beeflow.common.crt.charliecloud_driver import CharliecloudDriver
except ModuleNotFoundError:
    pass
try:
    from beeflow.common.crt.singularity_driver import SingularityDriver
except ModuleNotFoundError:
    pass


class WorkerError(Exception):
    """Worker error class."""

    def __init__(self, *args):
        """Construct a new worker error."""
        super(Exception, self).__init__()
        self.args = args


class Worker(ABC):
    """Worker interface for a generic workload manager."""

    def __init__(self, bee_workdir, **kwargs):
        """Load appropriate container runtime driver, based on configs in kwargs."""
        try:
            self.tm_crt = kwargs['container_runtime']
            if self.tm_crt == 'Charliecloud':
                crt_driver = CharliecloudDriver
            elif self.tm_crt == 'Singularity':
                crt_driver = SingularityDriver
            self.crt = ContainerRuntimeInterface(crt_driver)
        except KeyError:
            log.warning("No container runtime specified in config, proceeding with caution.")
            self.tm_crt = None
            crt_driver = None

        # Get BEE workdir from config file
        self.workdir = bee_workdir

    def task_save_path(self, task):
        """Return the task save path used for storing submission scripts output logs."""
        return f'{self.workdir}/workflows/{task.workflow_id}/{task.name}-{task.id}'

    def prepare(self, task):
        """Prepare for the task; create the task save directory, etc."""
        task_save_path = self.task_save_path(task)
        os.makedirs(task_save_path, exist_ok=True)

    @abstractmethod
    def build_text(self, task):
        """Build text for task script.

        :param task: task that we're building a script for
        :type task: Task
        :rtype: string
        """

    @abstractmethod
    def submit_task(self, task):
        """Worker submits task; returns job_id, job_state.

        :param task: instance of Task
        :rtype: tuple (int, string)
        """

    @abstractmethod
    def cancel_task(self, job_id):
        """Cancel task with job_id; returns job_state.

        :param job_id: to be cancelled
        :type job_id: integer
        :rtype: string
        """

    @abstractmethod
    def query_task(self, job_id):
        """Query job state for the task.

        :param job_id: job id to query for status.
        :type job_id: int
        :rtype: string
        """

# Ignore W0511: This allows us to have TODOs in the code
# pylama:ignore=W0511



"""Start up the workflow manager connecting all of the endpoints."""

import os
from flask import Flask
from celery import Celery # noqa (pylama can't find celery imports)
from beeflow.common.api import BeeApi
from beeflow.common import paths
from beeflow.wf_manager.resources.wf_list import WFList
from beeflow.wf_manager.resources.wf_actions import WFActions
from beeflow.wf_manager.resources.wf_metadata import WFMetadata
from beeflow.wf_manager.resources.wf_update import WFUpdate
from beeflow.wf_manager.resources import wf_utils


def create_app():
    """Create flask app object and add REST endpoints."""
    app = Flask(__name__)
    api = BeeApi(app)

    # Add endpoints
    api.add_resource(WFList, '/bee_wfm/v1/jobs/')
    api.add_resource(WFActions, '/bee_wfm/v1/jobs/<string:wf_id>')
    api.add_resource(WFMetadata, '/bee_wfm/v1/jobs/<string:wf_id>/metadata')
    api.add_resource(WFUpdate, '/bee_wfm/v1/jobs/update/')

    # Initialize celery app
    celery_app = Celery(app.name)
    redis_socket = os.path.join(paths.redis_root(), paths.redis_sock_fname())
    celery_app.config_from_object({
        'broker_url': f'redis+socket://{redis_socket}',
        'result_backend': f'db+sqlite://{paths.celery_db()}',
        'task_serializer': 'pickle',
        'accept_content': ['application/json', 'application/x-python-serialize'],
    })
    celery_app.set_default()
    app.extensions['celery'] = celery_app
    return app


if __name__ == '__main__':
    flask_app = create_app()
    bee_workdir = wf_utils.get_bee_workdir()
    # handler = bee_logging.save_log(bee_workdir=bee_workdir, log=log, logfile='wf_manager.log')
    # wfm_listen_port = wf_utils.get_open_port()
    # wf_db.set_wfm_port(wfm_listen_port)
    # flask_app.run(debug=False, port=str(wfm_listen_port))



#!/usr/bin/env python3
"""REST Interface for the BEE Scheduler."""

import argparse
import os

from flask import Flask, request
from flask_restful import Resource, Api

from beeflow.scheduler import algorithms
from beeflow.scheduler import task
from beeflow.scheduler import resource_allocation
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.db import sched_db
from beeflow.common import log as bee_logging

from beeflow.common.db.bdb import connect_db

log = bee_logging.setup(__name__)

flask_app = Flask(__name__)
api = Api(flask_app)

bee_workdir = bc.get('DEFAULT', 'bee_workdir')
db_path = bee_workdir + '/' + 'sched.db'


class ResourcesHandler(Resource):
    """Resources handler."""

    @staticmethod
    def put():
        """Create a list of resources to use for allocation."""
        db = connect_db(sched_db, db_path)
        db.resources.clear()
        resources = [resource_allocation.Resource.decode(r)
                     for r in request.json]
        db.resources.extend(resources)
        return f'created {len(resources)} resource(s)'

    @staticmethod
    def get():
        """Get a list of all resources."""
        db = connect_db(sched_db, db_path)
        return [r.encode() for r in db.resources]


class WorkflowJobHandler(Resource):
    """Schedule jobs for a specific workflow with the current resources."""

    @staticmethod
    def put(workflow_name):  # noqa ('workflow_name' may be used in the future)
        """Schedules a new list of independent tasks with available resources."""
        db = connect_db(sched_db, db_path)
        data = request.json
        tasks = [task.Task.decode(t) for t in data]
        # Pick the scheduling algorithm
        algorithm = algorithms.choose(**vars(flask_app.sched_conf))
        algorithm.schedule_all(tasks, list(db.resources))
        return [t.encode() for t in tasks]


api.add_resource(ResourcesHandler, '/bee_sched/v1/resources')
api.add_resource(WorkflowJobHandler,
                 '/bee_sched/v1/workflows/<string:workflow_name>/jobs')

# Default config values
SCHEDULER_PORT = 5100
# TODO: Use MODEL_FILE when interacting with MARS scheduling
MODEL_FILE = 'model'
ALLOC_LOGFILE = 'schedule_trace.txt'
MARS_CNT = 4
DEFAULT_ALGORITHM = 'fcfs'


def load_config_values():
    """Load the config, if necessary, and return config values.

    Load the config, if necessary, and return config values.
    """
    # Set the default config values
    conf = {
        'algorithm': None,
        'default_algorithm': None,
    }

    for key in conf:
        conf[key] = bc.get('scheduler', key)
    # Set some defaults
    conf['log'] = '/'.join([bee_workdir, 'logs', 'scheduler.log'])
    conf['workdir'] = os.path.join(bee_workdir, 'scheduler')
    conf['alloc_logfile'] = os.path.join(conf['workdir'], ALLOC_LOGFILE)

    conf = argparse.Namespace(**conf)
    log.info('Config = [')
    log.info(f'\talloc_logfile = {conf.alloc_logfile}')  # noqa pylama is wrong here
    log.info(f'\talgorithm = {conf.algorithm}')
    log.info(f'\tdefault_algorithm = {conf.default_algorithm}')
    log.info(f'\tworkdir = {conf.workdir}')  # noqa
    log.info(']')
    return conf


def create_app():
    """Create the Flask app for the scheduler."""
    # TODO: Refactor this to actually create the app here
    conf = load_config_values()
    flask_app.sched_conf = conf
    # Load algorithm data
    algorithms.load(**vars(conf))

    # Create the scheduler workdir, if necessary
    # sched_listen_port = wf_utils.get_open_port()
    # wf_db.set_sched_port(sched_listen_port)
    os.makedirs(conf.workdir, exist_ok=True) # noqa
    return flask_app

# Ignore W0511: This allows us to have TODOs in the code
# pylama:ignore=W0511



"""Tab completion code for client terminal input."""
import readline
import contextlib


@contextlib.contextmanager
def filepath_completion():
    """Tab complete files and pathnames within a context."""
    old_delims = readline.get_completer_delims()
    readline.set_completer_delims(' \n\t')
    readline.parse_and_bind('tab: complete')
    try:
        yield
    finally:
        # Reset the completer
        readline.set_completer_delims(old_delims)
        readline.parse_and_bind('')



"""Specialized functions for validation and config initialization."""
import os
from pathlib import Path


def validate_path(path):
    """Check that the path exists."""
    path = os.path.expanduser(path)
    if not os.path.exists(path):
        raise ValueError(f'path "{path}" does not exist')
    return path


def dir_(path):
    """Check that the path exists and is a directory."""
    path = validate_path(path)
    if not os.path.isdir(path):
        raise ValueError('path "{path}" is not a directory')
    return path


def make_dir(path):
    """Check if the dir exists and if not create it."""
    Path(path).mkdir(parents=True, exist_ok=True)
    return path


def parent_dir(path):
    """Ensure that the parent dir of path exists, or create it."""
    make_dir(Path(path).parent)
    return path


def file_(path):
    """Check that the path exists and is a file."""
    path = validate_path(path)
    if not os.path.isfile(path):
        raise ValueError(f'path "{path}" is not a file')
    return path


def nonnegative_int(value):
    """Validate that the input is nonnegative."""
    i = int(value)
    if i < 0:
        raise ValueError('the value must be nonnegative')
    return i


# NOTE: You must use validate_bool for all boolean values (since just using
#       bool, as in bool('False'), gives True for any string of length > 0)
def bool_(value):
    """Validate a boolean value."""
    return str(value).lower() == 'true'


def time_limit(value):
    """Validate a time limit entry."""
    value = value.strip()
    if not value:
        return ''
    _ = [int(part) for part in value.split(':')]
    return value



#! /usr/bin/env python3
"""Unit test module for the BEE CWL parser module."""

from pathlib import Path
import unittest
from beeflow.common.parser import CwlParser
from beeflow.common.wf_data import (generate_workflow_id, Workflow, Task, Hint,
                                    StepInput, StepOutput, InputParameter, OutputParameter)


REPO_PATH = Path(*Path(__file__).parts[:-3])


def find(path):
    """Find a path relative to the root of the repo."""
    return str(Path(REPO_PATH, path))


class TestParser(unittest.TestCase):
    """Unit test case for the workflow interface."""

    @classmethod
    def setUpClass(cls):
        """Initialize the CWL parser."""
        cls.parser = CwlParser()

    def test_parse_workflow_yaml(self):
        """Test parsing of workflow with a YAML input job file."""
        cwl_wf_file = find("examples/clamr-ffmpeg-build/clamr_wf.cwl")
        cwl_job_yaml = find("examples/clamr-ffmpeg-build/clamr_job.yml")
        workflow_id = generate_workflow_id()
        workflow, tasks = self.parser.parse_workflow(workflow_id, cwl_wf_file, cwl_job_yaml)

        self.assertEqual(workflow, WORKFLOW_GOLD)
        self.assertListEqual(tasks, TASKS_GOLD)
        for task in tasks:
            self.assertEqual(task.workflow_id, workflow_id)

    def test_parse_workflow_script(self):
        """Test parsing of workflow with a YAML input job file."""
        cwl_wf_file = find("beeflow/data/cwl/bee_workflows/clamr-ffmpeg-build_script/clamr_wf.cwl") #noqa
        cwl_job_yaml = find("beeflow/data/cwl/bee_workflows/clamr-ffmpeg-build_script/clamr_job.yml") #noqa

        workflow_id = generate_workflow_id()

        workflow, tasks = self.parser.parse_workflow(workflow_id, cwl_wf_file, cwl_job_yaml)

        self.assertEqual(workflow, WORKFLOW_GOLD)
        self.assertListEqual(tasks, TASKS_GOLD_SCRIPT)
        for task in tasks:
            self.assertEqual(task.workflow_id, workflow_id)

    def test_parse_workflow_json(self):
        """Test parsing of workflow with a JSON input job file."""
        cwl_wf_file = find("examples/clamr-ffmpeg-build/clamr_wf.cwl")
        cwl_job_json = find("examples/clamr-ffmpeg-build/clamr_job.json")
        workflow_id = generate_workflow_id()

        workflow, tasks = self.parser.parse_workflow(workflow_id, cwl_wf_file, cwl_job_json)

        self.assertEqual(workflow, WORKFLOW_GOLD)
        self.assertListEqual(tasks, TASKS_GOLD)
        for task in tasks:
            self.assertEqual(task.workflow_id, workflow_id)

    def test_parse_workflow_no_job(self):
        """Test parsing of a workflow without an input job file."""
        cwl_wf_file = find("beeflow/tests/cf.cwl")
        workflow_id = generate_workflow_id()
        # cwl_wf_file = "examples/clamr-ffmpeg-build/clamr_wf.cwl"

        workflow, tasks = self.parser.parse_workflow(workflow_id, cwl_wf_file)

        self.assertEqual(workflow, WORKFLOW_NOJOB_GOLD)
        self.assertListEqual(tasks, TASKS_NOJOB_GOLD)
        for task in tasks:
            self.assertEqual(task.workflow_id, workflow_id)


WORKFLOW_GOLD_SCRIPT = Workflow(
    name='clamr_wf',
    hints=[],
    requirements=[],
    inputs={InputParameter(id='input_format', type='string', value='image2'),
            InputParameter(id='time_steps', type='int', value=5000),
            InputParameter(id='output_filename', type='string', value='./CLAMR_movie.mp4'),
            InputParameter(id='frame_size', type='string', value='800x800'),
            InputParameter(id='frame_rate', type='int', value=12),
            InputParameter(id='max_levels', type='int', value=3),
            InputParameter(id='graphics_type', type='string', value='png'),
            InputParameter(id='steps_between_outputs', type='int', value=10),
            InputParameter(id='pixel_format', type='string', value='yuv420p'),
            InputParameter(id='grid_resolution', type='int', value=32),
            InputParameter(id='steps_between_graphics', type='int', value=25)},
    outputs={OutputParameter(id='clamr_stdout', type='File', value=None,
                             source='clamr/clamr_stdout'),
             OutputParameter(id='clamr_movie', type='File', value=None, source='ffmpeg/movie'),
             OutputParameter(id='ffmpeg_stderr', type='File', value=None,
                             source='ffmpeg/ffmpeg_stderr'),
             OutputParameter(id='clamr_time_log', type='File', value=None,
                             source='clamr/time_log')},
    workflow_id=generate_workflow_id())


WORKFLOW_GOLD = Workflow(
    name='clamr_wf',
    hints=[],
    requirements=[],
    inputs={InputParameter(id='input_format', type='string', value='image2'),
            InputParameter(id='time_steps', type='int', value=5000),
            InputParameter(id='output_filename', type='string', value='CLAMR_movie.mp4'),
            InputParameter(id='frame_size', type='string', value='800x800'),
            InputParameter(id='frame_rate', type='int', value=12),
            InputParameter(id='max_levels', type='int', value=3),
            InputParameter(id='graphics_type', type='string', value='png'),
            InputParameter(id='steps_between_outputs', type='int', value=10),
            InputParameter(id='pixel_format', type='string', value='yuv420p'),
            InputParameter(id='grid_resolution', type='int', value=32),
            InputParameter(id='steps_between_graphics', type='int', value=25)},
    outputs={OutputParameter(id='clamr_stdout', type='File', value=None,
                             source='clamr/clamr_stdout'),
             OutputParameter(id='clamr_movie', type='File', value=None, source='ffmpeg/movie'),
             OutputParameter(id='ffmpeg_stderr', type='File', value=None,
                             source='ffmpeg/ffmpeg_stderr'),
             OutputParameter(id='clamr_time_log', type='File', value=None,
                             source='clamr/time_log')},
    workflow_id=generate_workflow_id())

TASKS_GOLD_SCRIPT = [
    Task(
        name='clamr',
        base_command='/CLAMR/clamr_cpuonly',
        hints=[Hint(class_='DockerRequirement', params={'dockerFile': '# Dockerfile.clamr-ffmpeg\n# Developed on Chicoma @lanl\n# Patricia Grubel <pagrubel@lanl.gov>\n\nFROM debian:11\n\n\nRUN apt-get update && \\\n    apt-get install -y wget gnupg git cmake ffmpeg g++ make openmpi-bin libopenmpi-dev libpng-dev libpng16-16 libpng-tools imagemagick libmagickwand-6.q16-6 libmagickwand-6.q16-dev\n\nRUN git clone https://github.com/lanl/CLAMR.git\nRUN cd CLAMR && cmake . && make clamr_cpuonly\n', 'beeflow:containerName': 'clamr-ffmpeg'}), #noqa
               Hint(class_='beeflow:ScriptRequirement', params={'enabled': True, 'pre_script': 'echo "Before run"\n', 'post_script': 'echo "After run"\n'})], #noqa
        requirements=[],
        inputs=[StepInput(id='graphic_steps', type='int', value=None, default=None,
                          source='steps_between_graphics', prefix='-g', position=None,
                          value_from=None),
                StepInput(id='graphics_type', type='string', value=None, default=None,
                          source='graphics_type', prefix='-G', position=None, value_from=None),
                StepInput(id='grid_res', type='int', value=None, default=None,
                          source='grid_resolution', prefix='-n', position=None, value_from=None),
                StepInput(id='max_levels', type='int', value=None, default=None,
                          source='max_levels', prefix='-l', position=None, value_from=None),
                StepInput(id='output_steps', type='int', value=None, default=None,
                          source='steps_between_outputs', prefix='-i', position=None,
                          value_from=None),
                StepInput(id='time_steps', type='int', value=None, default=None,
                          source='time_steps', prefix='-t', position=None, value_from=None)],
        outputs=[StepOutput(id='clamr/clamr_stdout', type='stdout', value=None,
                            glob='clamr_stdout.txt'),
                 StepOutput(id='clamr/outdir', type='Directory', value=None,
                            glob='graphics_output/graph%05d.png'),
                 StepOutput(id='clamr/time_log', type='File', value=None,
                            glob='total_execution_time.log')],
        stdout='clamr_stdout.txt',
        stderr=None,
        workflow_id=WORKFLOW_GOLD.id
    ),
    Task(
        name='ffmpeg',
        base_command='ffmpeg -y',
        hints=[Hint(class_='DockerRequirement', params={'dockerFile': '# Dockerfile.clamr-ffmpeg\n# Developed on Chicoma @lanl\n# Patricia Grubel <pagrubel@lanl.gov>\n\nFROM debian:11\n\n\nRUN apt-get update && \\\n    apt-get install -y wget gnupg git cmake ffmpeg g++ make openmpi-bin libopenmpi-dev libpng-dev libpng16-16 libpng-tools imagemagick libmagickwand-6.q16-6 libmagickwand-6.q16-dev\n\nRUN git clone https://github.com/lanl/CLAMR.git\nRUN cd CLAMR && cmake . && make clamr_cpuonly\n', 'beeflow:containerName': 'clamr-ffmpeg'})], # noqa
        requirements=[],
        inputs=[StepInput(id='ffmpeg_input', type='Directory', value=None, default=None,
                          source='clamr/outdir', prefix='-i', position=2,
                          value_from='$("/graph%05d.png")'),
                StepInput(id='frame_rate', type='int', value=None, default=None,
                          source='frame_rate', prefix='-r', position=3, value_from=None),
                StepInput(id='frame_size', type='string', value=None, default=None,
                          source='frame_size', prefix='-s', position=4, value_from=None),
                StepInput(id='input_format', type='string', value=None, default=None,
                          source='input_format', prefix='-f', position=1, value_from=None),
                StepInput(id='output_file', type='string', value=None, default=None,
                          source='output_filename', prefix=None, position=6, value_from=None),
                StepInput(id='pixel_format', type='string', value=None, default=None,
                          source='pixel_format', prefix='-pix_fmt', position=5, value_from=None)],
        outputs=[StepOutput(id='ffmpeg/movie', type='File', value=None,
                            glob='$(inputs.output_file)'),
                 StepOutput(id='ffmpeg/ffmpeg_stderr', type='stderr', value=None,
                            glob='ffmpeg_stderr.txt')],
        stdout=None,
        stderr='ffmpeg_stderr.txt',
        workflow_id=WORKFLOW_GOLD.id)
]


TASKS_GOLD = [
    Task(
        name='clamr',
        base_command='/CLAMR/clamr_cpuonly',
        hints=[Hint(class_='DockerRequirement', params={'dockerFile': '# Dockerfile.clamr-ffmpeg\n# Developed on Chicoma @lanl\n# Patricia Grubel <pagrubel@lanl.gov>\n\nFROM debian:11\n\n\nRUN apt-get update && \\\n    apt-get install -y wget gnupg git cmake ffmpeg g++ make openmpi-bin libopenmpi-dev libpng-dev libpng16-16 libpng-tools imagemagick libmagickwand-6.q16-6 libmagickwand-6.q16-dev\n\nRUN git clone https://github.com/lanl/CLAMR.git\nRUN cd CLAMR && cmake . && make clamr_cpuonly\n', 'beeflow:containerName': 'clamr-ffmpeg'})], # noqa
        requirements=[],
        inputs=[StepInput(id='graphic_steps', type='int', value=None, default=None,
                          source='steps_between_graphics', prefix='-g', position=None,
                          value_from=None),
                StepInput(id='graphics_type', type='string', value=None, default=None,
                          source='graphics_type', prefix='-G', position=None, value_from=None),
                StepInput(id='grid_res', type='int', value=None, default=None,
                          source='grid_resolution', prefix='-n', position=None, value_from=None),
                StepInput(id='max_levels', type='int', value=None, default=None,
                          source='max_levels', prefix='-l', position=None, value_from=None),
                StepInput(id='output_steps', type='int', value=None, default=None,
                          source='steps_between_outputs', prefix='-i', position=None,
                          value_from=None),
                StepInput(id='time_steps', type='int', value=None, default=None,
                          source='time_steps', prefix='-t', position=None, value_from=None)],
        outputs=[StepOutput(id='clamr/clamr_stdout', type='stdout', value=None,
                            glob='clamr_stdout.txt'),
                 StepOutput(id='clamr/outdir', type='Directory', value=None,
                            glob='graphics_output/graph%05d.png'),
                 StepOutput(id='clamr/time_log', type='File', value=None,
                            glob='total_execution_time.log')],
        stdout='clamr_stdout.txt',
        stderr=None,
        workflow_id=WORKFLOW_GOLD.id
    ),
    Task(
        name='ffmpeg',
        base_command='ffmpeg -y',
        hints=[Hint(class_='DockerRequirement', params={'dockerFile': '# Dockerfile.clamr-ffmpeg\n# Developed on Chicoma @lanl\n# Patricia Grubel <pagrubel@lanl.gov>\n\nFROM debian:11\n\n\nRUN apt-get update && \\\n    apt-get install -y wget gnupg git cmake ffmpeg g++ make openmpi-bin libopenmpi-dev libpng-dev libpng16-16 libpng-tools imagemagick libmagickwand-6.q16-6 libmagickwand-6.q16-dev\n\nRUN git clone https://github.com/lanl/CLAMR.git\nRUN cd CLAMR && cmake . && make clamr_cpuonly\n', 'beeflow:containerName': 'clamr-ffmpeg'})], # noqa
        requirements=[],
        inputs=[StepInput(id='ffmpeg_input', type='Directory', value=None, default=None,
                          source='clamr/outdir', prefix='-i', position=2,
                          value_from='$("/graph%05d.png")'),
                StepInput(id='frame_rate', type='int', value=None, default=None,
                          source='frame_rate', prefix='-r', position=3, value_from=None),
                StepInput(id='frame_size', type='string', value=None, default=None,
                          source='frame_size', prefix='-s', position=4, value_from=None),
                StepInput(id='input_format', type='string', value=None, default=None,
                          source='input_format', prefix='-f', position=1, value_from=None),
                StepInput(id='output_file', type='string', value=None, default=None,
                          source='output_filename', prefix=None, position=6, value_from=None),
                StepInput(id='pixel_format', type='string', value=None, default=None,
                          source='pixel_format', prefix='-pix_fmt', position=5, value_from=None)],
        outputs=[StepOutput(id='ffmpeg/movie', type='File', value=None,
                            glob='$(inputs.output_file)'),
                 StepOutput(id='ffmpeg/ffmpeg_stderr', type='stderr', value=None,
                            glob='ffmpeg_stderr.txt')],
        stdout=None,
        stderr='ffmpeg_stderr.txt',
        workflow_id=WORKFLOW_GOLD.id)
]


WORKFLOW_NOJOB_GOLD = Workflow(
    name='cf',
    hints=[],
    requirements=[],
    inputs={InputParameter(id='infile', type='File', value=None)},
    outputs={OutputParameter(id='ffmpeg_movie', type='File', value=None, source='ffmpeg/outfile'),
             OutputParameter(id='clamr_dir', type='File', value=None, source='clamr/outfile')},
    workflow_id=generate_workflow_id())


TASKS_NOJOB_GOLD = [
    Task(
        name='clamr',
        base_command='/clamr/CLAMR-master/clamr_cpuonly -n 32 -l 3 -t 5000 -i 10 -g 25 -G png',
        hints=[Hint(class_='DockerRequirement',
                    params={'dockerImageId': '/usr/projects/beedev/clamr/clamr-toss.tar.gz'})],
        requirements=[],
        inputs=[StepInput(id='infile', type='File', value=None, default='lorem.txt',
                          source='infile', prefix=None, position=1, value_from=None)],
        outputs=[StepOutput(id='clamr/outfile', type='stdout', value=None,
                            glob='graphics_output')],
        stdout='graphics_output',
        stderr=None,
        workflow_id=WORKFLOW_NOJOB_GOLD.id),
    Task(
        name='ffmpeg',
        base_command='ffmpeg -f image2 -i $HOME/graphics_output/graph%05d.png -r 12 -s 800x800 -pix_fmt yuv420p $HOME/CLAMR_movie.mp4', # noqa
        hints=[],
        requirements=[],
        inputs=[StepInput(id='infile', type='File', value=None, default='graphics_output',
                          source='clamr/outfile', prefix=None, position=1, value_from=None)],
        outputs=[StepOutput(id='ffmpeg/outfile', type='stdout', value=None,
                            glob='CLAMR_movie.mp4')],
        stdout='CLAMR_movie.mp4',
        stderr=None,
        workflow_id=WORKFLOW_NOJOB_GOLD.id)
]


if __name__ == '__main__':
    unittest.main()



"""Slurm worker for work load management.

Builds command for submitting batch job.
"""

import io
import subprocess
import json
import urllib
import getpass
import requests_unixsocket
import requests

from beeflow.common import log as bee_logging
from beeflow.common.worker.worker import (Worker, WorkerError)
from beeflow.common import validation

log = bee_logging.setup(__name__)


class BaseSlurmWorker(Worker):
    """Base slurm worker code."""

    def __init__(self, default_account='', default_time_limit='', default_partition='', **kwargs):
        """Initialize the base slurm worker."""
        super().__init__(**kwargs)
        self.default_account = default_account
        self.default_time_limit = default_time_limit
        self.default_partition = default_partition

    def build_text(self, task):
        """Build text for task script."""
        task_save_path = self.task_save_path(task)
        crt_res = self.crt.run_text(task)
        stdout_param = ['--output', task.stdout]
        stderr_param = ['--error', task.stderr]
        if task.stdout and task.stderr:
            main_command_srun_args = stdout_param + stderr_param
        elif task.stdout:
            main_command_srun_args = stdout_param
        elif task.stderr:
            main_command_srun_args = stderr_param
        else:
            main_command_srun_args = []
        nodes = task.get_requirement('beeflow:MPIRequirement', 'nodes', default=1)
        ntasks = task.get_requirement('beeflow:MPIRequirement', 'ntasks', default=nodes)
        mpi_version = task.get_requirement('beeflow:MPIRequirement', 'mpiVersion', default='pmi2')
        time_limit = task.get_requirement('beeflow:SchedulerRequirement', 'timeLimit',
                                          default=self.default_time_limit)
        time_limit = validation.time_limit(time_limit)
        account = task.get_requirement('beeflow:SchedulerRequirement', 'account',
                                       default=self.default_account)
        partition = task.get_requirement('beeflow:SchedulerRequirement',
                                         'partition',
                                         default=self.default_partition)

        scripts_enabled = task.get_requirement('beeflow:ScriptRequirement', 'enabled',
                                               default=False)
        if scripts_enabled:
            # We use StringIO here to properly break the script up into lines with readlines
            pre_script = io.StringIO(task.get_requirement('beeflow:ScriptRequirement',
                                     'pre_script')).readlines()
            post_script = io.StringIO(task.get_requirement('beeflow:ScriptRequirement',
                                      'post_script')).readlines()
        # sbatch header
        script = [
            '#!/bin/bash',
            f'#SBATCH --job-name={task.name}-{task.id}',
            f'#SBATCH --output={task_save_path}/{task.name}-{task.id}.out',
            f'#SBATCH --error={task_save_path}/{task.name}-{task.id}.err',
            f'#SBATCH -N {nodes}',
            f'#SBATCH -n {ntasks}',
            '#SBATCH --open-mode=append',
        ]
        if time_limit:
            script.append(f'#SBATCH --time={time_limit}')
        if account:
            script.append(f'#SBATCH -A {account}')
        if partition:
            script.append(f'#SBATCH -p {partition}')

        # Return immediately on error
        script.append('set -e')
        script.append(crt_res.env_code)

        def srun(script_lines, script_cmd):
            """Wrap a pre or post command with srun."""
            cmd_args = ' '.join(script_cmd.args)
            if script_cmd.type == 'one-per-node':
                script.append(f'srun -N {nodes} -n {nodes} {cmd_args}')
            else:
                script_lines.append(f'srun {cmd_args}')

        # Pre commands
        if scripts_enabled:
            for cmd in pre_script:
                script.append(cmd)

        for cmd in crt_res.pre_commands:
            srun(script, cmd)

        # Main command
        srun_args = ' '.join(main_command_srun_args)
        args = ' '.join(crt_res.main_command.args)
        script.append(f'srun --mpi={mpi_version} {srun_args} {args}')

        # Post commands
        for cmd in crt_res.post_commands:
            srun(script, cmd)

        if scripts_enabled:
            for cmd in post_script:
                script.append(cmd)

        return '\n'.join(script)

    def write_script(self, task):
        """Build task script; returns filename of script."""
        task_text = self.build_text(task)

        task_script = f'{self.task_save_path(task)}/{task.name}-{task.id}.sh'
        with open(task_script, 'w', encoding='UTF-8') as script_f:
            script_f.write(task_text)
            script_f.close()
        return task_script

    def submit_job(self, script):
        """Worker submits job-returns (job_id, job_state)."""
        res = subprocess.run(['sbatch', '--parsable', script], text=True,  # noqa if we use check=True here, then we can't see stderr
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if res.returncode != 0:
            raise WorkerError(f'Failed to submit job: {res.stderr}')
        job_id = int(res.stdout)
        job_state = self.query_task(job_id)
        return job_id, job_state

    def submit_task(self, task):
        """Worker builds & submits script."""
        task_script = self.write_script(task)
        job_id, job_state = self.submit_job(task_script)
        return job_id, job_state


class SlurmrestdWorker(BaseSlurmWorker):
    """Worker class for when slurmrestd is available."""

    def __init__(self, bee_workdir, openapi_version, **kwargs):
        """Create a new Slurmrestd Worker object."""
        super().__init__(bee_workdir=bee_workdir, **kwargs)
        # Pull slurm socket configs from kwargs (Uses getpass.getuser() instead
        # of os.getlogin() because of an issue with using getlogin() without a
        # controlling terminal)
        self.slurm_socket = kwargs.get('slurm_socket', f'/tmp/slurm_{getpass.getuser()}.sock')
        self.session = requests_unixsocket.Session()
        encoded_path = urllib.parse.quote(self.slurm_socket, safe="")
        # Note: Socket path is encoded, http request is not generally.
        self.slurm_url = f"http+unix://{encoded_path}/slurm/{openapi_version}"

    def query_task(self, job_id):
        """Worker queries job; returns job_state."""
        try:
            resp = self.session.get(f'{self.slurm_url}/job/{job_id}')

            if resp.status_code != 200:
                raise WorkerError(f'Failed to query job {job_id}')
            data = json.loads(resp.text)
            # Check for errors in the response
            check_slurm_error(data, f'Failed to query job {job_id}')
            # For some versions of slurm, the job_state isn't included on failure
            try:
                job_state = data['jobs'][0]['job_state']
            except (KeyError, IndexError) as exc:
                raise WorkerError(f'Failed to query job {job_id}') from exc
        except requests.exceptions.ConnectionError:
            job_state = "NOT_RESPONDING"
        return job_state

    def cancel_task(self, job_id):
        """Worker cancels job, returns job_state."""
        try:
            resp = self.session.delete(f'{self.slurm_url}/job/{job_id}')
        except requests.exceptions.ConnectionError:
            return 'NOT_RESPONDING'
        # For some reason, some versions of slurmrestd are not returning an
        # HTTP error code, but just an error in the body
        errmsg = f'Unable to cancel job id {job_id}!'
        if resp.status_code != 200:
            raise WorkerError(f'{errmsg}: Bad response code: {resp.status_code}')
        try:
            data = resp.json()
            check_slurm_error(data, errmsg)
        except requests.exceptions.JSONDecodeError as exc:  # noqa requests is not installed in CI
            raise WorkerError(errmsg) from exc
        job_state = "CANCELLED"
        return job_state


class SlurmCLIWorker(BaseSlurmWorker):
    """Slurm worker interface that uses the CLI."""

    def query_task(self, job_id):
        """Query job state for the task."""
        # Use scontrol since it gives a lot of useful info
        try:
            res = subprocess.run(['scontrol', 'show', 'job', str(job_id)],
                                 text=True, check=True, stdout=subprocess.PIPE)
        except subprocess.CalledProcessError:
            raise WorkerError(
                f'Failed to query job {job_id}'
            ) from None

        def parse_key_val(pair):
            """Parse the key-value pair."""
            i = pair.find('=')
            return (pair[:i], pair[i + 1:])

        # Output is in a space-separated list of 'key=value' pairs
        pairs = res.stdout.split()
        key_vals = dict(parse_key_val(pair) for pair in pairs)
        return key_vals['JobState']

    def cancel_task(self, job_id):
        """Cancel task with job_id; returns job_state."""
        try:
            subprocess.run(['scancel', str(job_id)], text=True, check=True)
        except subprocess.CalledProcessError:
            raise WorkerError(f'Failed to cancel job {job_id}') from None
        return 'CANCELLED'


class SlurmWorker(Worker):
    """Main slurm worker class."""

    def __init__(self, use_commands, **kwargs):
        """Construct the slurm worker.

        :param use_commands: whether or not to use Slurm's CLI
        :type use_commands: bool
        """
        super().__init__(**kwargs)
        if use_commands:
            self._inner = SlurmCLIWorker(**kwargs)
        else:
            self._inner = SlurmrestdWorker(**kwargs)

    def build_text(self, task):
        """Build text for task script; use template if it exists."""
        return self._inner.build_text(task)

    def submit_task(self, task):
        """Worker submits task; returns job_id, job_state."""
        return self._inner.submit_task(task)

    def cancel_task(self, job_id):
        """Cancel task with job_id; returns job_state."""
        return self._inner.cancel_task(job_id)

    def query_task(self, job_id):
        """Query job state for the task."""
        return self._inner.query_task(job_id)


def check_slurm_error(data, msg):
    """Check for an error in a Slurm response."""
    if 'errors' in data and data['errors']:
        err = data['errors'][0]
        desc = err['description']
        errmsg = err['error']
        raise WorkerError(f'{msg}: {errmsg} ({desc})')



"""Graph the output of a PENNANT workflow."""
import re
import sys
import matplotlib.pyplot as plt


results = []
for fname in sys.argv[1:]:
    pe_count = 0
    times = []
    with open(fname, encoding='utf-8') as fp:
        for line in fp:
            # Check for the PE count
            m_pe_count = re.match(r'Running on (\d+) MPI PE\(s\)', line)
            if m_pe_count:
                pe_count = int(m_pe_count.group(1))
                continue
            # Check for an End cyle line
            if not line.startswith('End cycle'):
                continue
            _, _, _, wall = line.split(',')
            _, time = wall.split('=')
            time = float(time.strip())
            times.append(time)
    results.append({
        'pe_count': pe_count,
        'average_wall_time': sum(times) / len(times),
    })

# The node counts
x = [str(result['pe_count']) for result in results]
# Average wall for cycle
y = [result['average_wall_time'] for result in results]
fig, ax = plt.subplots()
ax.plot(x, y)
ax.set_title('PENNANT Workflow Run')
ax.set_xlabel('Node count')
ax.set_ylabel('Average wall time for cycle')
# Save to a png file
fig.savefig('graph.png')

# Ignore C0103: This is just a simple script, not all globals should be UPPER_CASE here
# pylama:ignore=C0103



"""Contains functions for managing a database for workflow and task information."""

import sqlite3
from sqlite3 import Error


def connect_db(module, db_path):
    """Return a DB object."""
    db = module.open_db(db_path)
    return db


def create_connection(db_file):
    """Create a new connection with the workflow database."""
    conn = None
    try:
        conn = sqlite3.connect(db_file)
        return conn
    except Error as error:
        print(error)
    return conn


def create_table(db_file, stmt):
    """Create a new table in the database."""
    with create_connection(db_file) as conn:
        try:
            cursor = conn.cursor()
            cursor.execute(stmt)
        except Error as error:
            print(error)


def run(db_file, stmt, params=None):
    """Run the sql statement on the database. Doesn't return anything."""
    with create_connection(db_file) as conn:
        try:
            cursor = conn.cursor()
            if params:
                cursor.execute(stmt, params)
            else:
                cursor.execute(stmt)
            conn.commit()
        except Error as error:
            print(error)


def getone(db_file, stmt, params=None):
    """Run the sql statement on the database and return the result."""
    with create_connection(db_file) as conn:
        try:
            cursor = conn.cursor()
            if params:
                cursor.execute(stmt, params)
            else:
                cursor.execute(stmt)
            result = cursor.fetchone()
        except Error:
            result = None
        return result


def getall(db_file, stmt, params=None):
    """Run the sql statement on the database and return the result."""
    with create_connection(db_file) as conn:
        try:
            cursor = conn.cursor()
            if params:
                cursor.execute(stmt, params)
            else:
                cursor.execute(stmt)
            result = cursor.fetchall()
        except Error:
            result = None
    return result


def table_exists(db_file, table_name):
    """Return true if a table exists and false if not."""
    stmt = f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}';"
    result = getall(db_file, stmt)
    return len(result) != 0


def get_table_length(db_file, table):
    """Return the number of rows in a table."""
    stmt = f"SELECT COUNT(*) from {table}"
    result = getall(db_file, stmt)
    rows = result[0][0]
    return rows



"""Tests of the Slurm worker."""
import uuid
import shutil
import time
import subprocess
import os
import pytest
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.worker_interface import WorkerInterface
from beeflow.common.worker.worker import WorkerError
from beeflow.common.worker.slurm_worker import SlurmWorker
from beeflow.common.wf_data import Task


# Timeout (seconds) for waiting on tasks
TIMEOUT = 150
# Extra slurmrestd arguments. This may be something to take on the command line
OPENAPI_VERSION = bc.get('slurm', 'openapi_version')
GOOD_TASK = Task(name='good-task', base_command=['sleep', '3'], hints=[],
                 requirements=[], inputs=[], outputs=[], stdout='', stderr='',
                 workflow_id=uuid.uuid4().hex)
BAD_TASK = Task(name='bad-task', base_command=['/this/is/not/a/command'], hints=[],
                requirements=[], inputs=[], outputs=[], stdout='', stderr='',
                workflow_id=uuid.uuid4().hex)


def wait_state(worker_iface, job_id, state):
    """Wait for Slurm to switch the job to another state."""
    time.sleep(1)
    n = 1
    last_state = worker_iface.query_task(job_id)
    while last_state == state:
        if n >= TIMEOUT:
            raise RuntimeError(f'job timed out, still in state {state}')
        time.sleep(1)
        n += 1
        last_state = worker_iface.query_task(job_id)
    return last_state


@pytest.fixture(params=[True, False], ids=['slurm-commands', 'slurmrestd'])
def slurm_worker(request):
    """Slurm worker fixture."""
    slurm_socket = f'/tmp/{uuid.uuid4().hex}.sock'
    bee_workdir = os.path.expanduser(f'/tmp/{uuid.uuid4().hex}.tmp')
    os.mkdir(bee_workdir)
    proc = subprocess.Popen(f'slurmrestd -s openapi/{OPENAPI_VERSION} unix:{slurm_socket}',
                            shell=True)
    time.sleep(1)
    worker_iface = WorkerInterface(worker=SlurmWorker, container_runtime='Charliecloud',
                                   slurm_socket=slurm_socket, bee_workdir=bee_workdir,
                                   openapi_version=OPENAPI_VERSION,
                                   use_commands=request.param)
    yield worker_iface
    time.sleep(1)
    proc.kill()
    shutil.rmtree(bee_workdir)


@pytest.fixture
def slurmrestd_worker_no_daemon():
    """Fixture that creates the worker interface but not slurmrestd."""
    slurm_socket = f'/tmp/{uuid.uuid4().hex}.sock'
    bee_workdir = os.path.expanduser(f'/tmp/{uuid.uuid4().hex}.tmp')
    os.mkdir(bee_workdir)
    yield WorkerInterface(worker=SlurmWorker, container_runtime='Charliecloud',
                          slurm_socket=slurm_socket, bee_workdir=bee_workdir,
                          openapi_version=OPENAPI_VERSION,
                          use_commands=False)
    shutil.rmtree(bee_workdir)


def test_good_task(slurm_worker):
    """Test submission of a good task."""
    job_id, last_state = slurm_worker.submit_task(GOOD_TASK)
    if last_state == 'PENDING':
        last_state = wait_state(slurm_worker, job_id, 'PENDING')
    if last_state == 'RUNNING':
        last_state = wait_state(slurm_worker, job_id, 'RUNNING')
    if last_state == 'COMPLETING':
        last_state = wait_state(slurm_worker, job_id, 'COMPLETING')
    assert last_state == 'COMPLETED'


def test_bad_task(slurm_worker):
    """Test submission of a bad task."""
    job_id, last_state = slurm_worker.submit_task(BAD_TASK)
    if last_state == 'PENDING':
        last_state = wait_state(slurm_worker, job_id, 'PENDING')
    if last_state == 'RUNNING':
        last_state = wait_state(slurm_worker, job_id, 'RUNNING')
    if last_state == 'COMPLETING':
        last_state = wait_state(slurm_worker, job_id, 'COMPLETING')
    assert last_state == 'FAILED'


def test_query_bad_job_id(slurm_worker):
    """Test querying a bad job ID."""
    with pytest.raises(WorkerError):
        slurm_worker.query_task(888)


def test_cancel_good_job(slurm_worker):
    """Cancel a good job."""
    job_id, _ = slurm_worker.submit_task(GOOD_TASK)
    job_state = slurm_worker.cancel_task(job_id)
    assert job_state == 'CANCELLED'


def test_no_slurmrestd(slurmrestd_worker_no_daemon):
    """Test without running slurmrestd."""
    worker = slurmrestd_worker_no_daemon
    job_id, state = worker.submit_task(GOOD_TASK)
    assert state == 'NOT_RESPONDING'
    assert worker.query_task(job_id) == 'NOT_RESPONDING'
    assert worker.cancel_task(job_id) == 'NOT_RESPONDING'
# Ignoring R1732: This is not what we need to do with the Popen of slurmrestd above;
#                 using a with statement doesn't kill the process immediately but just
#                 waits for it to complete and slurmrestd never will unless we kill it.
# Ignoring E402: "module level import not at top of file" - this is required for
#                bee config
# Ignoring W0621: Redefinition of names is required for pytest
# pylama:ignore=R1732,E402,W0621



"""Task Manager database code."""

from collections import namedtuple
import jsonpickle

from beeflow.common.db import bdb
from beeflow.common import log as bee_logging

log = bee_logging.setup(__name__)


class SubmitQueue:
    """Task Manager submit queue."""

    def __init__(self, db_file):
        """Construct a submit queue handler."""
        self.db_file = db_file
        self.Job = namedtuple("Job", "id task") #noqa Snake Case

    def __iter__(self):
        """Create an iterator for going over all elements."""
        stmt = 'SELECT task FROM submit_queue ORDER BY id ASC'
        result = bdb.getall(self.db_file, stmt)
        for rslt in result:
            task = jsonpickle.decode(rslt[0])
            yield task

    def count(self):
        """Count the number of items in the submit queue."""
        stmt = 'SELECT COUNT(*) AS count FROM submit_queue'
        count = bdb.getone(self.db_file, stmt)[0]
        return count

    def push(self, task):
        """Push the task onto the submit queue."""
        task_data = jsonpickle.encode(task)
        stmt = 'INSERT INTO submit_queue (task) VALUES (?)'
        bdb.run(self.db_file, stmt, [task_data])

    def pop(self):
        """Pop the bottom element off the queue."""
        select_stmt = 'SELECT id, task FROM submit_queue ORDER BY id ASC'
        result = bdb.getone(self.db_file, select_stmt)
        job = self.Job(*result)
        id_ = job.id
        task_data = job.task
        task = jsonpickle.decode(task_data)
        delete_stmt = 'DELETE FROM submit_queue WHERE id=?'
        bdb.run(self.db_file, delete_stmt, [id_])
        return task

    def clear(self):
        """Clear the submit queue."""
        stmt = 'DELETE FROM submit_queue'
        bdb.run(self.db_file, stmt)


class JobQueue:
    """Task Manager job queue."""

    def __init__(self, db_file):
        """Construct a job queue handler."""
        self.db_file = db_file
        self.Job = namedtuple("Task", "id task job_id job_state") # noqa Snake Case

    def __iter__(self):
        """Create an iterator for going over all elements in the queue."""
        stmt = 'SELECT id, task, job_id, job_state FROM job_queue ORDER BY id ASC'
        result = bdb.getall(self.db_file, stmt)
        for j in result:
            id_ = j[0]
            task = jsonpickle.decode(j[1])
            job_id = j[2]
            state = j[3]
            job = self.Job(id_, task, job_id, state)
            yield job

    def count(self):
        """Count the number of items in the job queue."""
        stmt = 'SELECT COUNT(*) AS count FROM job_queue'
        count = bdb.getone(self.db_file, stmt)[0]
        return count

    def push(self, task, job_id, job_state):
        """Push the job info onto the queue."""
        task_data = jsonpickle.encode(task)
        stmt = """INSERT INTO job_queue (task, job_id, job_state)
               VALUES (?, ?, ?)"""
        bdb.run(self.db_file, stmt, [task_data, job_id, job_state])

    def pop(self):
        """Pop the bottom element off the queue."""
        stmt = 'SELECT id, task, job_id, job_state FROM job_queue ORDER BY id ASC'
        result = bdb.getone(self.db_file, stmt)
        id_ = result[0]
        task = jsonpickle.decode(result[1])
        job_id = result[2]
        state = result[3]
        job = self.Job(id_, task, job_id, state)
        bdb.run(self.db_file, 'DELETE FROM job_queue WHERE id=?', [id_])
        return job

    def update_job_state(self, id_, job_state):
        """Update the job_state."""
        stmt = 'UPDATE job_queue SET job_state=? WHERE id=?'
        bdb.run(self.db_file, stmt, [job_state, id_])

    def remove_by_id(self, id_):
        """Remove a job from the queue by ID."""
        stmt = 'DELETE FROM job_queue WHERE id=?'
        bdb.run(self.db_file, stmt, [id_])

    def clear(self):
        """Clear the job queue."""
        stmt = 'DELETE FROM job_queue'
        bdb.run(self.db_file, stmt)


class TMDB:
    """Task Manager Database."""

    def __init__(self, db_file):
        """Construct a new TM database connection."""
        self.db_file = db_file
        self._init_tables()

    def _init_tables(self):
        """Initialize the workflow tables."""
        submit_queue_stmt = """CREATE TABLE IF NOT EXISTS submit_queue(
                        id INTEGER PRIMARY KEY ASC,
                        task TEXT)"""

        job_queue_stmt = """CREATE TABLE IF NOT EXISTS job_queue(
                        id INTEGER PRIMARY KEY ASC,
                        task TEXT,
                        job_id INTEGER,
                        job_state TEXT)"""

        bdb.create_table(self.db_file, submit_queue_stmt)
        bdb.create_table(self.db_file, job_queue_stmt)

    @property
    def submit_queue(self):
        """Return a SubmitQueue object."""
        return SubmitQueue(self.db_file)

    @property
    def job_queue(self):
        """Return a JobQueue object."""
        return JobQueue(self.db_file)


def open_db(db_file):
    """Open and return a new database."""
    return TMDB(db_file)



"""This module contains the workflow action endpoints."""
import shutil
import os

from flask import make_response, jsonify
from flask_restful import Resource, reqparse
from beeflow.common import log as bee_logging
from beeflow.wf_manager.resources import wf_utils

from beeflow.common.db import wfm_db
from beeflow.common.db.bdb import connect_db
from beeflow.wf_manager.common import dep_manager

log = bee_logging.setup(__name__)
db_path = wf_utils.get_db_path()


class WFActions(Resource):
    """Class to perform actions on existing workflows."""

    def __init__(self):
        """Initialize with passed json object."""
        self.reqparse = reqparse.RequestParser()

    def post(self, wf_id):
        """Start workflow. Send ready tasks to the task manager."""
        db = connect_db(wfm_db, db_path)
        if wf_utils.start_workflow(wf_id):
            db.workflows.update_workflow_state(wf_id, 'Running')
            resp = make_response(jsonify(msg='Started workflow!', status='ok'), 200)
        else:
            resp_body = jsonify(msg='Cannot start workflow it is {state.lower()}.', status='ok')
            resp = make_response(resp_body, 200)
        return resp

    @staticmethod
    def get(wf_id):
        """Check the database for the current status of all tasks."""
        db = connect_db(wfm_db, db_path)
        tasks = db.workflows.get_tasks(wf_id)
        tasks_status = []
        if not tasks:
            log.info(f"Bad query for wf {wf_id}.")
            wf_status = 'No workflow with that ID is currently loaded'
            resp = make_response(jsonify(tasks_status=tasks_status,
                                 wf_status=wf_status, status='not found'), 404)

        for task in tasks:
            tasks_status.append((task.id, task.name, task.state))
        wf_status = db.workflows.get_workflow_state(wf_id)

        resp = make_response(jsonify(tasks_status=tasks_status,
                             wf_status=wf_status, status='ok'), 200)
        return resp

    def delete(self, wf_id):
        """Cancel or delete the workflow. For cancel, current tasks finish running."""
        self.reqparse.add_argument('option', type=str, location='json')
        option = self.reqparse.parse_args()['option']
        db = connect_db(wfm_db, db_path)
        if option == "cancel":
            wfi = wf_utils.get_workflow_interface(wf_id)
            # Remove all tasks currently in the database
            if wfi.workflow_loaded():
                wfi.finalize_workflow()
            wf_utils.update_wf_status(wf_id, 'Cancelled')
            db.workflows.update_workflow_state(wf_id, 'Cancelled')
            log.info("Workflow cancelled")
            log.info("Shutting down gdb")
            pid = db.workflows.get_gdb_pid(wf_id)
            dep_manager.kill_gdb(pid)
            resp = make_response(jsonify(status='Cancelled'), 202)
        elif option == "remove":
            log.info(f"Removing workflow {wf_id}.")
            db.workflows.delete_workflow(wf_id)
            resp = make_response(jsonify(status='Removed'), 202)
            bee_workdir = wf_utils.get_bee_workdir()
            workflow_dir = f"{bee_workdir}/workflows/{wf_id}"
            shutil.rmtree(workflow_dir, ignore_errors=True)
            archive_path = f"{bee_workdir}/archives/{wf_id}.tgz"
            if os.path.exists(archive_path):
                os.remove(archive_path)
        return resp

    def patch(self, wf_id):
        """Pause or resume workflow."""
        db = connect_db(wfm_db, db_path)
        self.reqparse.add_argument('option', type=str, location='json')
        option = self.reqparse.parse_args()['option']

        wfi = wf_utils.get_workflow_interface(wf_id)
        wf_state = wfi.get_workflow_state()
        if option == 'pause' and wf_state == 'RUNNING':
            wfi.pause_workflow()
            wf_utils.update_wf_status(wf_id, 'Paused')
            db.workflows.update_workflow_state(wf_id, 'Paused')
            log.info("Workflow Paused")
            resp = make_response(jsonify(status='Workflow Paused'), 200)
        elif option == 'resume' and wf_state == 'PAUSED':
            wfi.resume_workflow()
            tasks = wfi.get_ready_tasks()
            wf_utils.schedule_submit_tasks(wf_id, tasks)
            wf_utils.update_wf_status(wf_id, 'Running')
            db.workflows.update_workflow_state(wf_id, 'Running')
            log.info("Workflow Resumed")
            resp = make_response(jsonify(status='Workflow Resumed'), 200)
        else:
            resp_msg = f'Cannot {option} workflow. It is currently {wf_state.lower()}.'
            log.info(resp_msg)
            resp = make_response(jsonify(status=resp_msg), 200)
        return resp



"""Handle task submission."""
from flask import jsonify, make_response
from flask_restful import Resource, reqparse
import jsonpickle
from beeflow.common import log as bee_logging
from beeflow.task_manager import utils

log = bee_logging.setup(__name__)


class TaskSubmit(Resource):
    """WFM sends tasks to the task manager."""

    @staticmethod
    def post():
        """Receives task from WFM."""
        db = utils.connect_db()
        parser = reqparse.RequestParser()
        parser.add_argument('tasks', type=str, location='json')
        data = parser.parse_args()
        tasks = jsonpickle.decode(data['tasks'])
        for task in tasks:
            db.submit_queue.push(task)
            log.info(f"Added {task.name} task to the submit queue")
        resp = make_response(jsonify(msg='Tasks Added!', status='ok'), 200)
        return resp



"""Inline workflow generation code."""
from pathlib import Path
import os
import uuid
import yaml

from beeflow.common.integration import utils


BASE_CONTAINER = 'alpine'
SIMPLE_DOCKERFILE = """
# Dummy docker file that really doesn't do much
FROM alpine

# Install something that has minimal dependencies
RUN apk update && apk add bzip2
"""
# Dump the Dockerfile to a path that the later code can reference
DOCKER_FILE_PATH = os.path.join('/tmp', f'Dockerfile-{uuid.uuid4().hex}')


def yaml_dump(path, data):
    """Dump this data as a yaml file at path."""
    with open(path, 'w', encoding='utf-8') as fp:
        yaml.dump(data, fp, Dumper=yaml.CDumper)


def builder_workflow(output_path, docker_requirement, main_input):
    """Generate a base workflow to be used for testing the builder."""
    os.makedirs(output_path, exist_ok=True)
    main_cwl_file = 'workflow.cwl'
    main_cwl_data = {
        'cwlVersion': 'v1.0',
        'class': 'Workflow',
        'inputs': {
            'main_input': 'string',
        },
        'outputs': {
            'main_output': {
                'outputSource': 'step0/step_output',
                'type': 'File',
            },
        },
        'steps': {
            'step0': {
                'run': 'step0.cwl',
                'in': {
                    'step_input': 'main_input',
                },
                'out': ['step_output'],
                'hints': {
                    'DockerRequirement': docker_requirement,
                },
            }
        },
    }
    yaml_dump(os.path.join(output_path, main_cwl_file), main_cwl_data)

    step0_file = 'step0.cwl'
    step0_data = {
        'cwlVersion': 'v1.0',
        'class': 'CommandLineTool',
        'baseCommand': 'touch',
        'inputs': {
            'step_input': {
                'type': 'string',
                'inputBinding': {
                    'position': 1,
                },
            },
        },
        'outputs': {
            'step_output': {
                'type': 'stdout',
            }
        },
        'stdout': 'output.txt',
    }
    yaml_dump(os.path.join(output_path, step0_file), step0_data)

    job_file = 'job.yml'
    job_data = {
        'main_input': main_input,
    }
    yaml_dump(os.path.join(output_path, job_file), job_data)

    return (main_cwl_file, job_file)


def simple_workflow(output_path, fname):
    """Generate a simple workflow."""
    os.makedirs(output_path, exist_ok=True)

    task0_cwl = 'task0.cwl'
    main_cwl = 'main.cwl'
    main_cwl_file = str(Path(output_path, main_cwl))
    main_cwl_data = {
        'cwlVersion': 'v1.0',
        'class': 'Workflow',
        'requirements': {},
        'inputs': {
            'in_fname': 'string',
        },
        'outputs': {
            'out': {
                'type': 'File',
                'outputSource': 'task0/out',
            },
        },
        'steps': {
            'task0': {
                'run': task0_cwl,
                'in': {
                    'fname': 'in_fname',
                },
                'out': ['out'],
            },
        },
    }
    yaml_dump(main_cwl_file, main_cwl_data)

    task0_cwl_file = str(Path(output_path, task0_cwl))
    task0_data = {
        'cwlVersion': 'v1.0',
        'class': 'CommandLineTool',
        'baseCommand': 'touch',
        'inputs': {
            'fname': {
                'type': 'string',
                'inputBinding': {
                    'position': 1,
                },
            },
        },
        'stdout': 'touch.log',
        'outputs': {
            'out': {
                'type': 'stdout',
            }
        }
    }
    yaml_dump(task0_cwl_file, task0_data)

    job_yaml = 'job.yaml'
    job_file = str(Path(output_path, job_yaml))
    job_data = {
        'in_fname': fname,
    }
    yaml_dump(job_file, job_data)
    return (main_cwl, job_yaml)


def init():
    """Initialize files and the container runtime."""
    with open(DOCKER_FILE_PATH, 'w', encoding='utf-8') as docker_file_fp:
        docker_file_fp.write(SIMPLE_DOCKERFILE)

    # Remove an existing base container
    if BASE_CONTAINER in utils.ch_image_list():
        utils.ch_image_delete(BASE_CONTAINER)



"""Utility code for running the integration tests."""
from contextlib import contextmanager
from pathlib import Path
import os
import shutil
import subprocess
import time
import traceback
import uuid

from beeflow.client import bee_client


# Max time of each workflow
TIMEOUT = 90
INTEGRATION_TEST_DIR = os.path.expanduser('~/.beeflow-integration')


class CIError(Exception):
    """CI error class."""

    def __init__(self, *args):
        """Initialize the error with default args."""
        self.args = args


class Container:
    """Container CI class for building BEE containers."""

    def __init__(self, name, dockerfile, tarball):
        """BEE CI container constructor."""
        self.name = name
        self.dockerfile = dockerfile
        self.tarball = tarball
        self.done = False

    def build(self):
        """Build the containers and save them in the proper location."""
        if not self.done:
            try:
                subprocess.check_call(['ch-image', 'build', '-f', self.dockerfile, '-t', self.name,
                                       '--force', 'seccomp', os.path.dirname(self.dockerfile)])
                subprocess.check_call(['ch-convert', '-i', 'ch-image', '-o', 'tar', self.name,
                                       self.tarball])
            except subprocess.CalledProcessError as error:
                raise CIError(
                    f'container build: {error}'
                ) from None
            self.done = True


class Workflow:
    """Workflow CI class for interacting with BEE."""

    def __init__(self, name, path, main_cwl, job_file, workdir, containers):
        """Workflow constructor."""
        self.name = name
        self.path = path
        self.main_cwl = main_cwl
        self.job_file = job_file
        self.workdir = workdir
        self.containers = containers
        self.wf_id = None
        self.tarball = None

    def run(self):
        """Attempt to submit and start the workflow."""
        # Build all the containers first
        print('Building all containers')
        for ctr in self.containers:
            ctr.build()
        try:
            tarball_dir = Path(self.path).parent
            tarball = f'{Path(self.path).name}.tgz'
            self.tarball = tarball_dir / tarball
            print('Creating the workflow tarball', self.tarball)
            bee_client.package(Path(self.path), Path(tarball_dir))
            print('Submitting and starting workflow')
            self.wf_id = bee_client.submit(self.name, self.tarball, self.main_cwl,
                                           self.job_file, self.workdir, no_start=False)
        except bee_client.ClientError as error:
            raise CIError(*error.args) from error

    @property
    def running(self):
        """Check if the workflow is running or about to run."""
        return bee_client.query(self.wf_id)[0] in ('Initializing', 'Waiting', 'Running', 'Pending')

    @property
    def status(self):
        """Get the status of the workflow."""
        return bee_client.query(self.wf_id)[0]

    @property
    def task_states(self):
        """Get the task states of the workflow."""
        return bee_client.query(self.wf_id)[1]

    def cleanup(self):
        """Clean up any leftover workflow data."""
        # Remove the generated tarball
        os.remove(self.tarball)


class TestRunner:
    """Test runner class."""

    def __init__(self):
        """Build a new test runner class."""
        self.test_cases = []
        self.timeout = TIMEOUT

    def add(self, ignore=False):
        """Decorate a test case and add it to the runner instance."""

        def wrap(test_case):
            """Wrap the function."""
            # First turn it into a contextmanager
            test_case = contextmanager(test_case)
            self.test_cases.append((test_case, ignore))
            return test_case

        return wrap

    def test_details(self):
        """Return a list of all the test details (test_name, ignore)."""
        return [(test_case.__name__, ignore) for test_case, ignore in self.test_cases]

    def _run_workflows(self, workflows):
        """Run a list of workflows to completion."""
        # Start all workflows at once
        for wfl in workflows:
            wfl.run()
        # Now run until all workflows are complete or until self.timeout is hit
        t = 0
        while t < self.timeout:
            if all(not wfl.running for wfl in workflows):
                # All workflows have completed
                break
            time.sleep(2)
            t += 2
        if t >= self.timeout:
            raise CIError('workflow timeout')

    def run(self, test_names=None):
        """Test run all test cases."""
        results = {}
        for test_case, ignore in self.test_cases:
            if test_names is not None:
                # Skip tests that aren't required to be run
                if test_case.__name__ not in test_names:
                    continue
            elif ignore:
                # Skip ignored tests
                continue
            outer_workdir = os.path.join(INTEGRATION_TEST_DIR, uuid.uuid4().hex)
            os.makedirs(outer_workdir)
            msg0 = f'Starting test {test_case.__name__}'
            msg1 = f'(running in "{outer_workdir}")'
            count = max(len(msg0), len(msg1))
            print('#' * count)
            print(msg0)
            print(msg1)
            print('-' * count)
            try:
                with test_case(outer_workdir) as workflows:
                    self._run_workflows(workflows)
            except CIError as error:
                traceback.print_exc()
                results[test_case.__name__] = error
                print('------')
                print('FAILED')
                print('------')
            else:
                results[test_case.__name__] = None
                print('------')
                print('PASSED')
                print('------')
                # Only remove the outer_workdir if it passed
                try:
                    shutil.rmtree(outer_workdir)
                except OSError as err:
                    print(f'WARNING: Failed to remove {outer_workdir}')
                    print(err)

        fails = sum(1 if result is not None else 0 for _, result in results.items())
        self.display_results(results, fails)
        return 1 if fails > 0 else 0

    def display_results(self, results, fails):
        """Show the workflow results."""
        print('######## WORKFLOW RESULTS ########')
        passes = len(results) - fails
        print(f'{passes} passes, {fails} fails')
        for name, result in results.items():
            error = result
            if error is not None:
                print(f'{name}: {error}')
        print('##################################')


#
# Helper methods for running the test cases
#


def ch_image_delete(img_name):
    """Execute a ch-image delete [img_name]."""
    try:
        subprocess.check_call(['ch-image', 'delete', img_name])
    except subprocess.CalledProcessError:
        raise CIError(
            f'failed when calling `ch-image delete {img_name}` to clean up'
        ) from None


def ch_image_list():
    """Execute a ch-image list and return the images."""
    try:
        res = subprocess.run(['ch-image', 'list'], stdout=subprocess.PIPE, check=True)
        output = res.stdout.decode(encoding='utf-8')
        images = [img for img in output.split('\n') if img]
        return images
    except subprocess.CalledProcessError as err:
        raise CIError(f'failed when calling `ch-image list`: {err}') from err


def ci_assert(predicate, msg):
    """Assert that the predicate is True, or raise a CIError."""
    if not predicate:
        raise CIError(msg)


def check_path_exists(path):
    """Check that the specified path exists."""
    ci_assert(os.path.exists(path), f'expected file "{path}" does not exist')


def check_completed(workflow):
    """Ensure the workflow has a completed status."""
    ci_assert(workflow.status == 'Archived', f'bad workflow status {workflow.status}')


def check_workflow_failed(workflow):
    """Ensure that the workflow completed in a Failed state."""
    ci_assert(workflow.status == 'Failed',
              f'workflow did not fail as expected (final status: {workflow.status})')



"""Predict Code."""
import json
import pickle
import click
# import numpy as np
# from sklearn.linear_model import LinearRegression
# from sklearn.tree import DecisionTreeClassifier
# from sklearn.model_selection import train_test_split
# from sklearn import metrics


@click.command()
@click.argument('e', type=float)
@click.argument('i', type=float)
@click.argument('t', type=float)
def pred(e, i, t):
    """Predict."""
    # Import Linear Regression model
    regression_model = pickle.load(open('/home/bee/cwl2/mymodel.p', 'rb'))
    # Import Decision Tree classifier model
    dt_model = pickle.load(open('/home/bee/cwl2/clf1.p', 'rb'))
    predict_linear_regression = regression_model.predict([[e, i, t]])
    predict_linear_regression_list = predict_linear_regression.tolist()
    predict_linear_regression_json_str = json.dumps(predict_linear_regression_list)

    predict_decision_tree = dt_model.predict([[e, i, t]])
    predict_decision_tree_list = predict_decision_tree.tolist()
    predict_decision_tree_json_str = json.dumps(predict_decision_tree_list)

    print("Expected Salary from Regression is $", predict_linear_regression)
    print("Expected Salary from DT is $", predict_decision_tree)

    key = 'Predicted Salary (S) by Regression and DT Model'
    values = (predict_linear_regression_json_str, predict_decision_tree_json_str)
    click.echo(json.dumps({key: values}))


if __name__ == '__main__':
    pred(e=4, i=5, t=6)
# Ignores preserving code for now
# pylama:ignore=C0103,R1732,C0501



"""Decision Tree."""
# import json
import pickle
import click
# import numpy as np
# from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeClassifier
# from sklearn.model_selection import train_test_split
# from sklearn import metrics


@click.command()
@click.argument('x1', type=int)
def reg(x1):
    """Reqression."""
    # Import X and Y datasets
    X = pickle.load(open('/home/bee/cwl2/MyX.p', 'rb'))
    Y = pickle.load(open('/home/bee/cwl2/MyY.p', 'rb'))

    X = X.values

    print("My pickle X is", X)
    Y = Y.values
    print("My pickle Y is", Y)
    for i in range(x1):
        clf = DecisionTreeClassifier()
        clf1 = clf.fit(X, Y)
    print("Decision tree parameters")
    print(clf1)
    pickle.dump(clf1, open('clf1.p', 'wb'))


if __name__ == '__main__':
    reg(x1=1)
# Ignores preserving code for now
# pylama:ignore=C0103,R1732,W0612,W0621



"""Background task code.

This code processes submitted tasks, monitors status, and sends info back to
the Workflow Manager.
"""
import traceback
import jsonpickle
from beeflow.task_manager import utils
from beeflow.common import log as bee_logging
from beeflow.common.build.utils import ContainerBuildError
from beeflow.common.build_interfaces import build_main


log = bee_logging.setup(__name__)


def update_task_state(workflow_id, task_id, job_state, **kwargs):
    """Informs the workflow manager of the current state of a task."""
    data = {'wf_id': workflow_id, 'task_id': task_id, 'job_state': job_state}
    if 'metadata' in kwargs:
        kwargs['metadata'] = jsonpickle.encode(kwargs['metadata'])

    if 'task_info' in kwargs:
        kwargs['task_info'] = jsonpickle.encode(kwargs['task_info'])

    data.update(kwargs)
    conn = utils.wfm_conn()
    resp = conn.put(utils.wfm_resource_url("update/"), json=data)
    if resp.status_code != 200:
        log.warning("WFM not responding when sending task update.")


def resolve_environment(task):
    """Use build interface to create a valid environment.

    This will build and/or pull containers if necessary; it can take some time
    to run this step.
    """
    build_main(task)


def submit_task(db, worker, task):
    """Submit (or resubmit) a task."""
    try:
        log.info(f'Resolving environment for task {task.name}')
        resolve_environment(task)
        log.info(f'Environment preparation complete for task {task.name}')
        job_id, job_state = worker.submit_task(task)
        log.info(f'Job Submitted {task.name}: job_id: {job_id} job_state: {job_state}')
        # place job in queue to monitor
        db.job_queue.push(task=task, job_id=job_id, job_state=job_state)
        # update_task_metadata(task.id, task_metadata)
    except ContainerBuildError as err:
        job_state = 'BUILD_FAIL'
        log.error(f'Failed to build container for {task.name}: {err}')
        log.error(f'{task.name} state: {job_state}')
    except Exception as err:  # noqa (we have to catch everything here)
        # Set job state to failed
        job_state = 'SUBMIT_FAIL'
        log.error(f'Task Manager submit task {task.name} failed! \n {err}')
        log.error(f'{task.name} state: {job_state}')
        # Log the traceback information as well
        log.error(traceback.format_exc())
    # Send the initial state to WFM
    # update_task_state(task.id, job_state, metadata=task_metadata)
    return job_state


def submit_jobs():
    """Submit all jobs currently in submit queue to the workload scheduler."""
    db = utils.connect_db()
    worker = utils.worker_interface()
    while db.submit_queue.count() >= 1:
        # Single value dictionary
        task = db.submit_queue.pop()
        job_state = submit_task(db, worker, task)
        update_task_state(task.workflow_id, task.id, job_state)


def update_jobs():
    """Check and update states of jobs in queue, remove completed jobs."""
    db = utils.connect_db()
    worker = utils.worker_interface()
    # Need to make a copy first
    job_q = list(db.job_queue)
    for job in job_q:
        id_ = job.id
        task = job.task
        job_id = job.job_id
        job_state = job.job_state
        new_job_state = worker.query_task(job_id)

        # If state changes update the WFM
        if job_state != new_job_state:
            db.job_queue.update_job_state(id_, new_job_state)
            if new_job_state in ('FAILED', 'TIMELIMIT', 'TIMEOUT'):
                # Harvest lastest checkpoint file.
                task_checkpoint = task.get_full_requirement('beeflow:CheckpointRequirement')
                log.info(f'state: {new_job_state}')
                log.info(f'TIMELIMIT/TIMEOUT task_checkpoint: {task_checkpoint}')
                if task_checkpoint:
                    try:
                        checkpoint_file = utils.get_restart_file(task_checkpoint, task.workdir)
                        task_info = {'checkpoint_file': checkpoint_file, 'restart': True}
                        update_task_state(task.workflow_id, task.id, new_job_state,
                                          task_info=task_info)
                    except utils.CheckpointRestartError as err:
                        log.error(f'Checkpoint restart failed for {task.name} ({task.id}): {err}')
                        update_task_state(task.workflow_id, task.id, 'FAILED')
                else:
                    update_task_state(task.workflow_id, task.id, new_job_state)
            # States are based on https://slurm.schedmd.com/squeue.html#SECTION_JOB-STATE-CODES
            elif new_job_state in ('BOOT_FAIL', 'NODE_FAIL', 'OUT_OF_MEMORY', 'PREEMPTED'):
                # Don't update wfm, just resubmit
                log.info(f'Task {task.name} in state {new_job_state}')
                log.info(f'Resubmitting task {task.name}')
                db.job_queue.remove_by_id(id_)
                job_state = submit_task(db, worker, task)
                update_task_state(task.workflow_id, task.id, job_state)
            else:
                update_task_state(task.workflow_id, task.id, new_job_state)

        if job_state in ('ZOMBIE', 'COMPLETED', 'CANCELLED', 'FAILED', 'TIMEOUT', 'TIMELIMIT'):
            # Remove from the job queue. Our job is finished
            db.job_queue.remove_by_id(id_)


def process_queues():
    """Look for newly submitted jobs and update status of scheduled jobs."""
    submit_jobs()
    update_jobs()



"""Workflow and Task state values."""


class WorkflowStates:
    """Workflow status values."""

    INITIALIZING = 'INITIALIZING'
    PENDING = 'PENDING'
    RUNNING = 'RUNNING'
    PAUSED = 'PAUSED'
    RESUME = 'RESUME'
    CANCELLED = 'CANCELLED'


class TaskStates:
    """Task status values."""

    PENDING = 'PENDING'
    RUNNING = 'RUNNING'
    COMPLETED = 'COMPLETED'
    FAILED = 'FAILED'
    PAUSED = 'PAUSED'
    CANCELLED = 'CANCELLED'



"""Centralized path data for everything stored in the workdir."""
import os
from beeflow.common.config_driver import BeeConfig as bc


def _workdir():
    """Return the workdir."""
    return bc.get('DEFAULT', 'bee_workdir')


def _sockdir():
    """Return the socket directory."""
    sockdir = os.path.join(_workdir(), 'sockets')
    os.makedirs(sockdir, exist_ok=True)
    return sockdir


def beeflow_socket():
    """Return the socket for the beeflow daemon."""
    return os.path.join(_sockdir(), 'beeflow.sock')


def wfm_socket():
    """Get the socket path for the Workflow Manager."""
    return os.path.join(_sockdir(), 'wf_manager.sock')


def tm_socket():
    """Get the socket path for the Task Manager."""
    return os.path.join(_sockdir(), 'task_manager.sock')


def sched_socket():
    """Get the scheduler socket."""
    return os.path.join(_sockdir(), 'scheduler.sock')


def slurm_socket():
    """Get the slurm socket (for slurmrestd)."""
    return os.path.join(_sockdir(), 'slurmrestd.sock')


def log_path():
    """Return the main log path."""
    bee_workdir = bc.get('DEFAULT', 'bee_workdir')
    return os.path.join(bee_workdir, 'logs')


def log_fname(component):
    """Determine the log file name for the given component."""
    return os.path.join(log_path(), f'{component}.log')


def redis_root():
    """Get the redis root directory (create it if it doesn't exist)."""
    path = os.path.join(_workdir(), 'redis')
    os.makedirs(path, exist_ok=True)
    return path


def redis_container():
    """Get the path to the unpacked Redis container."""
    return os.path.join(_workdir(), 'redis_container')


def redis_sock_fname():
    """Return the file name for the Redis socket."""
    return 'redis.sock'


def _celery_root():
    """Get the celery root directory (create it if it doesn't exist)."""
    path = os.path.join(_workdir(), 'celery')
    os.makedirs(path, exist_ok=True)
    return path


def celery_config():
    """Return the celery config path."""
    return os.path.join(_celery_root(), 'celery.py')


def celery_db():
    """Return the celery db path."""
    return os.path.join(_celery_root(), 'celery.db')
